import os
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

########################################################################
class LogManager(object):
    """
    Logs are the records written by the various components of Notebook server.
    You can query the logs and change various log settings.
    """
    _url = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<LogManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<LogManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    def clean(self):
        """
        Deletes all the log files on all server machines in the site. This is an irreversible
        operation.

        This operation forces the server to clean the logs, which has the effect of freeing
        up disk space. However, it is not required that you invoke this operation because
        the server periodically purges old logs.

        :return: boolean

        """
        params = {
            "f" : "json",
        }
        url = "{}/clean".format(self._url)
        res = self._con.post(url,
                             params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def settings(self):
        """
        Get/set the current log settings.

        :returns: PropertyMap
        """
        params = {
            "f" : "json"
        }
        url = self._url + "/settings"
        try:
            res = self._con.get(url, params)
            return PropertyMap(res)
        except:
            return ""
    #----------------------------------------------------------------------
    @settings.setter
    def settings(self, value):
        """
        Get/set the current log settings.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        value                  dict. A dictionary with the key/values pairs to modify settings.
        ==================     ====================================================================

        :returns: PropertyMap
        """
        params = {'f':'json'}
        current = dict(self.settings)
        allowed_keys = list(current.keys())
        for k in allowed_keys:
            if k in value:
                params[k] = value[k]
            else:
                params[k] = current[k]
        url = self._url + "/settings/edit"
        res = self._gis._con.post(url, params)
    #----------------------------------------------------------------------
    def query(self,
              start_time=None,
              end_time=None,
              since_server_start=False,
              level="WARNING",
              services="*",
              machines="*",
              server="*",
              codes=None,
              process_IDs=None,
              export=False,
              export_type="CSV", #CSV or TAB
              out_path=None):
        """
        The query operation on the logs resource provides a way to
        aggregate, filter, and page through logs across the entire site.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        start_time             Optional string. The most recent time to query.  Default is now.
                               Time can be specified in milliseconds since UNIX epoch, or as an
                               ArcGIS Server timestamp. For example { "startTime": "2011-08-01T15:17:20,123", ... },
                               { "startTime": 1312237040123, ... }, respectively.
        ------------------     --------------------------------------------------------------------
        end_time               Optional string. The oldest time to include in the result set. You
                               can use this to limit the query to the last n minutes or hours as
                               needed. Default is the beginning of all logging.
        ------------------     --------------------------------------------------------------------
        since_server_start     Optional string. Gets only the records written since the server
                               started (True).  The default is False.
        ------------------     --------------------------------------------------------------------
        level                  Optional string. Gets only the records with a log level at or more
                               severe than the level declared here. Can be one of (in severity
                               order): DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The
                               default is WARNING.
        ------------------     --------------------------------------------------------------------
        services               Optional string. Query records related to a specific service.
                               The default is all.
        ------------------     --------------------------------------------------------------------
        machines               Optional string. Query records related to a specific machine.
                               The default is all.
        ------------------     --------------------------------------------------------------------
        server                 Optional string. Query records related to a specific server.
                               The default is all.
        ------------------     --------------------------------------------------------------------
        codes                  Optional string. Gets only the records with the specified code.
                               The default is all.  See http://server.arcgis.com/en/server/latest/administer/windows/log-codes-overview.htm
        ------------------     --------------------------------------------------------------------
        process_IDs            Optional string. Query by the machine process ID that logged the event.
        ------------------     --------------------------------------------------------------------
        export                 Optional string. Boolean indicating whether to export the query
                               results.  The default is False (don't export).
        ------------------     --------------------------------------------------------------------
        export_type            Optional string. The export file type. CSV or TAB are the choices,
                               CSV is the default.
        ------------------     --------------------------------------------------------------------
        out_path               Optional string. The path to download the log file to.
        ==================     ====================================================================

        :return:
           A JSON of the log items that match the query. If export option is set to True, the
           output log file path is returned.


        """


        if codes is None:
            codes = []
        if process_IDs is None:
            process_IDs = []
        allowed_levels = ("SEVERE", "WARNING", "INFO",
                          "FINE", "VERBOSE", "DEBUG")
        qFilter = {
            "services": "*",
            "machines": "*",
            "server" : "*"
        }
        if len(process_IDs) > 0:
            qFilter['processIds'] = process_IDs
        if len(codes) > 0:
            qFilter['codes'] = codes
        params = {
            "f" : "json",
            "sinceServerStart" : since_server_start,
            "pageSize" : 10000
        }
        url = "{url}/query".format(url=self._url)
        if start_time is not None and \
           isinstance(start_time, datetime):
            params['startTime'] = start_time.strftime("%Y-%m-%dT%H:%M:%S")
        if end_time is not None and \
           isinstance(end_time, datetime):
            params['endTime'] = end_time.strftime("%Y-%m-%dT%H:%M:%S")
        if level.upper() in allowed_levels:
            params['level'] = level
        if server != "*":
            qFilter['server'] = server.split(',')
        if services != "*":
            qFilter['services'] = services.split(',')
        if machines != "*":
            qFilter['machines'] = machines.split(",")
        params['filter'] = qFilter
        if export is True and \
           out_path is not None:

            messages = self._con.get(url,
                                     params)
            with open(name=out_path, mode='wb') as f:
                hasKeys = False
                if export_type == "TAB":
                    csvwriter = csv.writer(f, delimiter='\t')
                else:
                    csvwriter = csv.writer(f)
                for message in messages['logMessages']:
                    if hasKeys == False:
                        csvwriter.writerow(message.keys())
                        hasKeys = True
                    csvwriter.writerow(message.values())
                    del message
            del messages
            return out_path
        else:
            return self._con.get(url,
                                 params)







"""
These functions calculate total counts, lengths, areas, and basic descriptive statistics of features and their attributes
within areas or near other features.

aggregate_points calculates statistics about points that fall within specified areas.
summarize_nearby calculates statistics for features and their attributes that are within a specified distance.
summarize_within calculates statistics for area features and attributes that overlap each other.
"""
import arcgis as _arcgis
from arcgis._impl.common._utils import _date_handler
import arcgis.network as network


def aggregate_points(point_layer,
                     polygon_layer,
                     keep_boundaries_with_no_points=True,
                     summary_fields=[],
                     group_by_field=None,
                     minority_majority=False,
                     percent_points=False,
                     output_name=None,
                     context=None,
                     gis=None,
                     estimate=False,
                     future=False):
    """
    .. image:: _static/images/agg_points_standard/aggregate_points.png

    The Aggregate Points task works with a layer of point features and a layer of polygon features. It first figures out which points fall within each polygon's area.
    After determining this point-in-polygon spatial relationship, statistics about all points in the polygon are calculated and assigned to the area. The most basic statistic is the count of the number of points within the polygon, but you can get other statistics as well.
    For example, if your points represented coffee shops and each point has a TOTAL_SALES attribute, you can get statistics like the sum of all TOTAL_SALES within the polygon, or the minimum or maximum TOTAL_SALES value, or the standard deviation of all sales within the polygon.

    ====================================     ====================================================================
    **Parameter**                            **Description**
    ------------------------------------     --------------------------------------------------------------------
    point_layer                              Required point layer. The point features that will be aggregated into the polygons in the polygon_layer. See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------     --------------------------------------------------------------------
    polygon_layer                            Required polygon layer. The polygon features (areas) into which the input points will be aggregated. See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------     --------------------------------------------------------------------
    keep_boundaries_with_no_points           Optional boolean. A Boolean value that specifies whether the polygons that have no points within them should be returned in the output. The default is true.
    ------------------------------------     --------------------------------------------------------------------
    summary_fields                           Optional list of strings. A list of field names and statistical summary type that you wish to calculate for all points within each polygon.
                                             Note that the count of points within each polygon is always returned.
                                             summary type is one of the following:

                                             * Sum - Adds the total value of all the points in each polygon
                                             * Mean - Calculates the average of all the points in each polygon.
                                             * Min - Finds the smallest value of all the points in each polygon.
                                             * Max - Finds the largest value of all the points in each polygon.
                                             * Stddev - Finds the standard deviation of all the points in each polygon.
                                             Example [fieldName1 summaryType1,fieldName2 summaryType2].
    ------------------------------------     --------------------------------------------------------------------
    group_by_field                           Optional string. A field name in the point_layer. Points that have the same value for the group by field will have their own counts and summary field statistics. You can create statistical groups using an attribute in the analysis layer. For example, if you are aggregating crimes to neighborhood boundaries, you may have an attribute Crime_type with five different crime types. Each unique crime type forms a group, and the statistics you choose will be calculated for each unique value of Crime_type. When you choose a grouping attribute, two results are created: the result layer and a related table containing the statistics.
    ------------------------------------     --------------------------------------------------------------------
    minority_majority                        Optional boolean. This boolean parameter is applicable only when a group_by_field is specified. If true, the minority (least dominant) or the majority (most dominant) attribute values for each group field within each boundary are calculated. Two new fields are added to the aggregated_layer prefixed with Majority_ and Minority_.
                                             The default is false.
    ------------------------------------     --------------------------------------------------------------------
    percent_points                           Optional boolean. This boolean parameter is applicable only when a group_by_field is specified. If set to true, the percentage count of points for each unique group_by_field value is calculated. A new field is added to the group summary output table containing the percentages of each attribute value within each group. If minority_majority is true, two additional fields are added to the aggregated_layer containing the percentages of the minority and majority attribute values within each group.
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional string. Output Features Name (str). Optional parameter.
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional string. Context contains additional settings that affect task execution. For Aggregate Points, there are two settings.

                                             #. Extent (extent)-a bounding box that defines the analysis area. Only those points in the input pointLayer that intersect the bounding box will be analyzed.
                                             #. Output Spatial Reference (outSR) the output features will be projected into the output spatial reference.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    estimate                                 Optional Boolean. If True, the number of credits to run the operation will be returned.
    ====================================     ====================================================================

    :returns: result_layer : feature layer Item if output_name is specified, else Feature Collection.


    .. code-block:: python

        USAGE EXAMPLE: To find number of permits issued in each zip code of US.

        agg_result = aggregate_points(point_layer=permits,
                                polygon_layer=zip_codes,
                                keep_boundaries_with_no_points=False,
                                summary_fields=["DeclValNu mean","DeclValNu2 mean"],
                                group_by_field='Declared_V',
                                minority_majority=True,
                                percent_points=True,
                                output_name="aggregated_permits",
                                context='{"extent":{"xmin":-8609738.077325115,"ymin":4743483.445485223,"xmax":-8594030.268012533,"ymax":4752206.821338257,"spatialReference":{"wkid":102100,"latestWkid":3857}}}')

    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.aggregate_points(
                     point_layer,
                     polygon_layer,
                     keep_boundaries_with_no_points,
                     summary_fields,
                     group_by_field,
                     minority_majority,
                     percent_points,
                     output_name,
                     context,
                     estimate=estimate, future=future)



def summarize_nearby(sum_nearby_layer,
                     summary_layer,
                     near_type="StraightLine",
                     distances=[],
                     units="Meters",
                     time_of_day=None,
                     time_zone_for_time_of_day="GeoLocal",
                     return_boundaries=True,
                     sum_shape=True,
                     shape_units=None,
                     summary_fields=[],
                     group_by_field=None,
                     minority_majority=False,
                     percent_shape=False,
                     output_name=None,
                     context=None,
                     gis=None,
                     estimate=False,
                     future=False):
    """
    .. image:: _static/images/summarize_nearby/summarize_nearby.png

    The ``summarize_nearby`` method finds features that are within a specified distance of features in the input layer.
    Distance can be measured as a straight-line distance, a drive-time distance (for example, within 10 minutes), or a
    drive distance (within 5 kilometers). Statistics are then calculated for the nearby features. For example:

    * Calculate the total population within five minutes of driving time of a proposed new store location.
    * Calculate the number of freeway access ramps within a one-mile driving distance of a proposed new store location to use as a measure of
      store accessibility.

    =========================    ====================================================================================================================
    **Parameter**                **Description**
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    sum_nearby_layer             Required feature layer. Point, line, or polygon features from which distances will be measured to features in the ``summary_layer``. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    summary_layer                Required layer. Point, line, or polygon features. Features in this layer that are within the specified distance to features in the ``sum_nearby_layer`` will be summarized. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    near_type                    Optional string. Defines what kind of distance measurement you want to use: straight-line distance, or by measuring travel
                                 time or travel distance along a street network using various modes of transportation known as travel modes.

                                 The default is 'StraightLine'.

                                 Choice list: ['StraightLine', 'Driving Distance', 'Driving Time', 'Rural Driving Distance', 'Rural Driving Time', 'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    distances                    Optional float. Float values that defines the search distance (for 'StraightLine' and distance based travel modes) or time (for time based travel modes).
                                 You can enter a single distance value or multiple values, separating each value with a space. Features that are within (or equal to) the distances you
                                 enter will be summarized. The units of the distance values is supplied by the units parameter.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    units                        Otional string. If ``near_type`` is 'StraightLine' or a distance-based travel mode, this is the linear unit to be used with the distance value(s) specified in distances.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']

                                 If ``near_type`` is a time based travel mode, the following values can be used as units:

                                 Choice list: ['Seconds', 'Minutes', 'Hours']

                                 The default is 'Meters'.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    time_of_day                  Optional datetime.datetime. Specify whether travel times should consider traffic conditions. To use traffic in the analysis,
                                 set ``near_type`` to a travel mode object whose impedance_attribute_name property is set to travel_time and assign a value
                                 to ``time_of_day``. (A travel mode with other impedance_attribute_name values don't support traffic.) The ``time_of_day`` value represents
                                 the time at which travel begins, or departs, from the origin points. The time is specified as datetime.datetime.

                                 The service supports two kinds of traffic: typical and live. Typical traffic references travel speeds that are made up of historical
                                 averages for each five-minute interval spanning a week. Live traffic retrieves speeds from a traffic feed that processes phone probe
                                 records, sensors, and other data sources to record actual travel speeds and predict speeds for the near future.

                                 The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries
                                 Esri currently provides traffic data for.

                                 Typical Traffic:

                                 To ensure the task uses typical traffic in locations where it is available, choose a time and day of the week, and then convert the day
                                 of the week to one of the following dates from 1990:

                                 * Monday - 1/1/1990
                                 * Tuesday - 1/2/1990
                                 * Wednesday - 1/3/1990
                                 * Thursday - 1/4/1990
                                 * Friday - 1/5/1990
                                 * Saturday - 1/6/1990
                                 * Sunday - 1/7/1990
                                 Set the time and date as datetime.datetime.

                                 For example, to solve for 1:03 p.m. on Thursdays, set the time and date to 1:03 p.m., 4 January 1990; and convert to
                                 datetime eg. datetime.datetime(1990, 1, 4, 1, 3).

                                 Live Traffic:

                                 To use live traffic when and where it is available, choose a time and date and convert to datetime.

                                 Esri saves live traffic data for 12 hours and references predictive data extending 12 hours into the future. If the time and date you
                                 specify for this parameter is outside the 24-hour time window, or the travel time in the analysis continues past the predictive data window,
                                 the task falls back to typical traffic speeds.

                                 Examples:
                                 from datetime import datetime

                                 * ``time_of_day``- datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                 * ``time_of_day``- datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                 * ``time_of_day``- datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014,
                                   live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                 * ``time_of_day``- datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015,
                                   live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.

    -------------------------    --------------------------------------------------------------------------------------------------------------------
    time_zone_for_time_of_day     Optional string. Specify the time zone or zones of the ``time_of_day`` parameter.

                                  Choice list: ['GeoLocal', 'UTC']

                                  GeoLocal-refers to the time zone in which the originsLayer points are located.

                                  UTC-refers to Coordinated Universal Time.

                                  The default is 'GeoLocal'.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    return_boundaries            Optional boolean. If true, the ``result_layer`` will contain areas defined by the specified ``near_type``. For example, if using 'StraightLine' of 5 miles,
                                 the ``result_layer`` will contain areas with a 5 mile radius around the input ``sum_nearby_layer`` features.

                                 If False, the ``result_ayer`` will contain the same features as the ``sum_nearby_layer``.

                                 The default is True.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    sum_shape                    Optional boolean. A boolean value that instructs the task to calculate statistics based on shape type of the ``summary_layer``,
                                 such as the length of lines or areas of polygons of the ``summary_layer`` within each polygon in ``sum_within_layer``.

                                 The default is True.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    shape_units                  Optional string. If ``sum_shape`` is true, you must specify the units of the shape summary.
                                 Values:

                                 When ``summary_layer`` contains polygons: Values: ['Acres', 'Hectares', 'SquareMeters', 'SquareKilometers', 'SquareFeet', 'SquareYards', 'SquareMiles']
                                 When ``summary_layer`` contains lines: Values: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    summary_fields               Optional list of strings.A list of field names and statistical summary types that you want to calculate.
                                 Note that the count is always returned by default.

                                 fieldName is the name of one of the numeric fields found in the input join layer.

                                 statisticType is one of the following:

                                 * ``SUM``-Adds the total value of all the points in each polygon
                                 * ``MEAN``-Calculates the average of all the points in each polygon
                                 * ``MIN``-Finds the smallest value of all the points in each polygon
                                 * ``MAX``-Finds the largest value of all the points in each polygon
                                 * ``STDDEV``-Finds the standard deviation of all the points in each polygon

                                 Example: ["fieldName summaryType","fieldName summaryType", ...]
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    group_by_field               Optional string. This is a field of the ``summary_layer`` features that you can use to calculate statistics separately for each unique attribute value.
                                 For example, suppose the ``summary_layer`` contains point locations of businesses that store hazardous materials, and one of the fields is HazardClass
                                 containing codes that describe the type of hazardous material stored. To calculate summaries by each unique value of HazardClass, use HazardClass as
                                 the ``group_by_field`` field.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    minority_majority            Optional boolean. This boolean parameter is applicable only when a ``group_by_field`` is specified. If true, the minority (least dominant) or the
                                 majority (most dominant) attribute values for each group field within each nearby area are calculated. Two new fields are added to
                                 the ``result_layer`` prefixed with Majority_ and Minority_.

                                 The default is False.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    percent_shape                Optional boolean. This Boolean parameter is applicable only when a ``group_by_field`` is specified. If set to true,
                                 the percentage of each unique ``group_by_field`` value is calculated for each ``sum_nearby_layer`` feature.

                                 The default is False.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    output_name                  Optional string. If provided, the task will create a feature service of the results. You define the name of the service.
                                 If ``output_name`` is not supplied, the task will return a feature collection.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    context                      Optional string. Context contains additional settings that affect task execution. For  ``summarize_nearby``, there are two settings.

                                 #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the ``sum_nearby_layer`` and ``summary_layer`` that intersect the bounding box will be analyzed.
                                 #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    estimate                     Optional boolean. Returns the number of credit for the operation.
    -------------------------    --------------------------------------------------------------------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =========================    ====================================================================================================================

    :returns: result_layer : feature layer Item if ``output_name`` is specified, else Feature Collection.

         dict with the following keys:

             "result_layer" : layer (FeatureCollection)

             "group_by_summary" : layer (FeatureCollection)

    .. code-block:: python

         # USAGE EXAMPLE: To find hospital facilities that are within 5 miles of a school.
         summarize_nearby(sum_nearby_layer=item2.layers[0],
                          summary_layer=item1.layers[0],
                          near_type='StraightLine',
                          distances=[5],
                          units='Miles',
                          time_zone_for_time_of_day='GeoLocal',
                          return_boundaries=False,
                          sum_shape=True,
                          shape_units=None,
                          output_name='nearest hospitals to schools')
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    if isinstance(near_type, str):
        if near_type != 'StraightLine':
            route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
            near_type = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == near_type][0]

    return gis._tools.featureanalysis.summarize_nearby(
                     sum_nearby_layer,
                     summary_layer,
                     near_type,
                     distances,
                     units,
                     _date_handler(time_of_day),
                     time_zone_for_time_of_day,
                     return_boundaries,
                     sum_shape,
                     shape_units,
                     summary_fields,
                     group_by_field,
                     minority_majority,
                     percent_shape,
                     output_name,
                     context,
                     estimate=estimate, future=future)


def summarize_center_and_dispersion(
        analysis_layer,
        summarize_type=["CentralFeature"],
        ellipse_size=None,
        weight_field=None,
        group_field=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):

    """
    .. image:: _static/images/summarize_center_and_dispersion/summarize_center_and_dispersion.png

    The ``summarize_center_and_dispersion`` method finds central features and directional distributions. It can be used to answer questions such as:

    * Where is the center?
    * Which feature is the most accessible from all other features?
    * How dispersed, compact, or integrated are the features?
    * Are there directional trends?s

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          Required frature layer. The point, line, or polygon features to be analyzed. See :ref:`Feature Input<FeatureInput>`.
    --------------------    ---------------------------------------------------------
    summarize_type          Required list of strings. The method with which to summarize the ``analysis_layer``.

                            Choice list: ["CentralFeature", "MeanCenter", "MedianCenter", "Ellipse"]
    --------------------    ---------------------------------------------------------
    ellipse_size            Optional string. The size of the output ellipse in standard deviations.

                            Choice list: ['1 standard deviations', '2 standard deviations', '3 standard deviations']

                            The default ellipse size is '1 standard deviations'.
    --------------------    ---------------------------------------------------------
    weight_field            Optional field. A numeric field in the ``analysis_layer`` to be used to
                            weight locations according to their relative importance.
    --------------------    ---------------------------------------------------------
    group_field             Optional field. The field used to group features for separate directional
                            distribution calculations. The ``group_field`` can be of
                            integer, date, or string type.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. If provided, the method will create a feature service of the results.
                            You define the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    --------------------    ---------------------------------------------------------
    context                 Optional string. Context contains additional settings that affect task execution. For ``summarize_center_and_dispersion``, there are two settings.

                            #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the input layer that intersect the bounding box will be buffered.
                            #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    --------------------    ---------------------------------------------------------
    estimate                Optional boolean. If True, the number of credits to run the operation will be returned.
    --------------------    ---------------------------------------------------------
    future                  Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================    =========================================================

    :returns: list of items if ``output_name`` is supplied else, a Python dictionary with the following keys:
        "central_feature_result_layer" : layer (FeatureCollection)
        "mean_feature_result_layer" : layer (FeatureCollection)
        "median_feature_result_layer" : layer (FeatureCollection)
        "ellipse_feature_result_layer" : layer (FeatureCollection)

    .. code-block:: python

        # USAGE EXAMPLE: To find central features and mean center of earthquake over past months.
        central_features = summarize_center_and_dispersion(analysis_layer=earthquakes,
                                                           summarize_type=["CentralFeature","MeanCenter"],
                                                           ellipse_size='2 standard deviations',
                                                           weight_field='mag',
                                                           group_field='magType',
                                                           output_name='find central features and mean center of earthquake over past months')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size,
        weight_field,
        group_field,
        output_name,
        context,
        estimate=estimate, future=future)


def summarize_within(sum_within_layer,
                     summary_layer,
                     sum_shape=True,
                     shape_units=None,
                     summary_fields=[],
                     group_by_field=None,
                     minority_majority=False,
                     percent_shape=False,
                     output_name=None,
                     context=None,
                     gis=None,
                     estimate=False,
                     future=False):
    """
    .. image:: _static/images/summarize_within/summarize_within.png

    The ``summarize_within`` method finds the point, line, or polygon features (or portions of these features)
    that are within the boundaries of polygons in another layer. For example:

        * Given a layer of watershed boundaries and a layer of land-use boundaries by land-use type, calculate total acreage of land-use type for each watershed.
        * Given a layer of parcels in a county and a layer of city boundaries, summarize the average value of vacant parcels within each city boundary.
        * Given a layer of counties and a layer of roads, summarize the total mileage of roads by road type within each county.

    You can think of ``summarize_within`` as taking two layers and stacking them on top of each other.
    One of the layers, the ``sum_within_layer`` must be a polygon layer, and imagine that these polygon
    boundaries are all colored red. The other layer, the ``summary_layer``, can be any feature type point,
    line, or polygon. After stacking these layers on top of each other, you peer down through the stack
    and count the number of features in the ``summary_layer`` that fall within the polygons with the red
    boundaries (the ``sum_within_layer``). Not only can you count the number of features, you can calculate
    simple statistics about the attributes of the features in the ``summary_layer``, such as sum, mean, minimum, maximum, and so on.

    =====================================    =========================================================
    **Argument**                             **Description**
    -------------------------------------    ---------------------------------------------------------
    sum_within_layer                         Required feature layer. The polygon features. Features, or
                                             portions of features, in the ``summary_layer`` (below) that fall within
                                             the boundaries of these polygons will be summarized. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    summary_layer                            Required feature layer. Point, line, or polygon features that will be summarized for each polygon in the ``sum_within_layer``.
                                             See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    sum_shape                                Optional boolean. A boolean value that instructs the task to calculate statistics
                                             based on shape type of the ``summary_layer``, such as the length of lines or areas of
                                             polygons of the ``summary_layer`` within each polygon in ``sum_within_layer``.

                                             The default is True.
    -------------------------------------    ---------------------------------------------------------
    shape_units                              Optional string. Specify units to summarize the length or areas when ``sum_shape`` is set to true. Units is not required to summarize
                                             points.

                                             When ``summary_layer`` contains polygons: ['Acres', 'Hectares', 'SquareMeters', 'SquareKilometers', 'SquareMiles', 'SquareYards', 'SquareFeet']

                                             When ``summary_layer`` contains lines: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------------------    ---------------------------------------------------------
    summary_fields                           Optional list of strings. A list of field names and statistical summary type that you wish
                                             to calculate for all features in the ``summary_layer`` that are within each polygon in the ``sum_within_layer`` .

                                             Example: ["fieldname1 summary", "fieldname2 summary"]
    -------------------------------------    ---------------------------------------------------------
    group_by_field                           Optional string. This is a field of the ``summary_layer`` features that you can use to calculate statistics separately
                                             for each unique attribute value. For example, suppose the ``sum_within_layer`` contains city boundaries and
                                             the ``summary_layer`` features are parcels. One of the fields of the parcels is Status which contains
                                             two values: VACANT and OCCUPIED. To calculate the total area of vacant and occupied parcels within the
                                             boundaries of cities, use Status as the ``group_by_field`` field.
    -------------------------------------    ---------------------------------------------------------
    minority_majority                        Optional boolean. This boolean parameter is applicable only when a ``group_by_field`` is specified.
                                             If true, the minority (least dominant) or the majority (most dominant) attribute values for each group
                                             field are calculated. Two new fields are added to the ``result_layer`` prefixed with Majority_ and Minority_.

                                             The default is False.
    -------------------------------------    ---------------------------------------------------------
    percent_shape                            Optional boolean. This Boolean parameter is applicable only when a ``group_by_field`` is specified.
                                             If set to true, the percentage of each unique ``group_by_field`` value is calculated for
                                             each ``sum_within_layer`` polygon.

                                             The default is False.
    -------------------------------------    ---------------------------------------------------------
    output_name                              Optional string. If provided, the method will create a feature service of the results.
                                             You define the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    -------------------------------------    ---------------------------------------------------------
    context                                  Optional string. Context contains additional settings that affect task execution. For ``summarize_within``, there are two settings.

                                             #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the ``sum_within_layer`` and the ``Summary_layer`` that intersect the bounding box will be summarized.
                                             #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------------------    ---------------------------------------------------------
    estimate                                 Optional boolean. If True, the number of credits to run the operation will be returned.
    -------------------------------------    ---------------------------------------------------------
    future                                   Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =====================================    =========================================================

    :returns: Item if ``output_name`` is set. else results in a Python dict with the following keys:

        dict with the following keys:

            "result_layer" : layer (FeatureCollection)

            "group_by_summary" : layer (FeatureCollection)

    .. code-block:: python

        # USAGE EXAMPLE: To summarize traffic accidents within each county and group them by the day of accident.
        acc_within_county = summarize_within(sum_within_layer=boundaries,
                                             summary_layer=collision_lyr,
                                             sum_shape=True,
                                             group_by_field='Day',
                                             minority_majority=True,
                                             percent_shape=True,
                                             output_name='summarize accidents within each county',
                                             context={"extent":{"xmin":-13160690.837046918,"ymin":4041586.5461609075,"xmax":-13132466.464352652,"ymax":4058001.397985127,"spatialReference":{"wkid":102100,"latestWkid":3857}}})
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.summarize_within(
                     sum_within_layer,
                     summary_layer,
                     sum_shape,
                     shape_units,
                     summary_fields,
                     group_by_field,
                     minority_majority,
                     percent_shape,
                     output_name,
                     context,
                     estimate=estimate, future=future)


def join_features(target_layer,
                  join_layer,
                  spatial_relationship=None,
                  spatial_relationship_distance=None,
                  spatial_relationship_distance_units=None,
                  attribute_relationship=None,
                  join_operation="""JoinOneToOne""",
                  summary_fields=None,
                  output_name=None,
                  context=None,
                  gis=None,
                  estimate=False,
                  future=False):
    """
    .. image:: _static/images/join_features/join_features.png

    The ``join_features`` method works with two layers and joins the attributes
    from one feature to another based on spatial and attribute relationships.

    ============================================================================================     =================================================================================================================================
    **Parameter**                                                                                    **Description**
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    target_layer                                                                                     Required layer. The point, line, polygon or table layer that will have attributes from
                                                                                                     the ``join_layer`` appended to its table. See :ref:`Feature Input<FeatureInput>`.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    join_layer                                                                                       Required layer. The point, line, polygon or table layer that will be joined to the ``target_layer``. See :ref:`Feature Input<FeatureInput>`.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    spatial_relationship                                                                             Required string. Defines the spatial relationship used to spatially join features.

                                                                                                     Choice list: ['identicalto', 'intersects', 'completelycontains', 'completelywithin', 'withindistance']
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    spatial_relationship_distance (Required if ``spatial_relationship`` is withindistance)           Optional float. A float value used for the search distance to determine if the target features are near or within a
                                                                                                     specified distance of the join features.
                                                                                                     This is only applied if Within a distance of is the selected ``spatial_relationship``.
                                                                                                     You can only enter a single distance value. The units of the distance values are supplied by the
                                                                                                     ``spatial_relationship_distance_units`` parameter.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    spatial_relationship_distance_units (Required if ``spatial_relationship`` is withindistance)     Optional string. The linear unit to be used with the distance value specified in ``spatial_relationship_distance``.

                                                                                                     Choice list: ['Miles', 'Yards', 'Feet', 'NauticalMiles', 'Meters', 'Kilometers']

                                                                                                     The default is 'Miles'.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    attribute_relationship                                                                           Optional list of dicts. Defines an attribute relationship used to join features. Features are matched when the field
                                                                                                     values in the join layer are equal to field values in the target layer.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    join_operation                                                                                   Optional string. A string representing the type of join that will be applied.

                                                                                                     Choice list: ['JoinOneToOne', 'JoinOneToMany']

                                                                                                        * ``JoinOneToOne`` - If multiple join features are found that have the same relationships with a
                                                                                                          single target feature, the attributes from the multiple join features will be aggregated using
                                                                                                          the specified summary statistics. For example, if a point target feature is found within two
                                                                                                          separate polygon join features, the attributes from the two polygons will be aggregated before
                                                                                                          being transferred to the output point feature class. If one polygon has an attribute value of
                                                                                                          3 and the other has a value of 7, and a SummaryField of sum is selected, the aggregated value
                                                                                                          in the output feature class will be 10. There will always be a Count field calculated, with a
                                                                                                          value of 2, for the number of features specified. This is the default.

                                                                                                        * ``JoinOneToMany`` - If multiple join features are found that have the same relationship with
                                                                                                          a single target feature, the output feature class will contain multiple copies (records) of
                                                                                                          the target feature. For example, if a single point target feature is found within two separate
                                                                                                          polygon join features, the output feature class will contain two copies of the target feature:
                                                                                                          one record with the attributes of the first polygon, and another record with the attributes of
                                                                                                          the second polygon. There are no summary statistics calculated with this method.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    summary_fields                                                                                   Optional list of dicts. A list of field names and statistical summary types that you want to calculate.
                                                                                                     Note that the count is always returned by default.

                                                                                                     fieldName is the name of one of the numeric fields found in the input join layer.

                                                                                                     statisticType is one of the following:

                                                                                                        * ``SUM`` - Adds the total value of all the points in each polygon
                                                                                                        * ``MEAN`` - Calculates the average of all the points in each polygon
                                                                                                        * ``MIN`` - Finds the smallest value of all the points in each polygon
                                                                                                        * ``MAX`` - Finds the largest value of all the points in each polygon
                                                                                                        * ``STDDEV`` - Finds the standard deviation of all the points in each polygon
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    output_name                                                                                      Optional string. If provided, the method will create a feature service of the results. You define the name of the service.
                                                                                                     If ``output_name`` is not supplied, the task will return a feature collection.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    context                                                                                          Optional string. Context contains additional settings that affect method execution. For ``join_features``, there are the following two settings:

                                                                                                     #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the input layer that intersect the bounding box will be analyzed.

                                                                                                     #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    estimate                                                                                         Optional boolean. If True, the number of credits to run the operation will be returned.
    --------------------------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------------------------------------
    future                                                                                           Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ============================================================================================     =================================================================================================================================

    :returns: result_layer : feature layer Item if ``output_name`` is specified, else feature collection.

    .. code-block:: python

        USAGE EXAMPLE: To summarize traffic accidents within each parcel using spatial relationship.
        accident_count_in_each_parcel = join_features(target_layer=parcel_lyr,
                                                      join_layer=traffic_accidents_lyr,
                                                      spatial_relationship='intersects',
                                                      output_name='join features',
                                                      context={"extent":{"xmin":-9375809.87305117,"ymin":4031882.3806860778,"xmax":-9370182.196843527,"ymax":4034872.9794178144,"spatialReference":{"wkid":102100,"latestWkid":3857}}}, )
    """
    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.join_features(
        target_layer,
        join_layer,
        spatial_relationship,
        spatial_relationship_distance,
        spatial_relationship_distance_units,
        attribute_relationship,
        join_operation,
        summary_fields,
        output_name,
        context,
        estimate=estimate, future=future)



"""
Generates Layer Types from the given inputs.

"""
from __future__ import absolute_import
import os
from six import add_metaclass
from six.moves.urllib_parse import urlparse
from arcgis.gis import GIS
from arcgis.features.layer import FeatureLayer, FeatureLayerCollection
from arcgis.geocoding import Geocoder
from arcgis.geoprocessing._tool import Toolbox
from arcgis._impl.tools import _GeometryService as GeometryService
from arcgis.network import NetworkDataset
from arcgis.gis import Layer
from arcgis.mapping import VectorTileLayer
from arcgis.mapping import MapImageLayer
from arcgis.raster import ImageryLayer
from arcgis.schematics import SchematicLayers
from arcgis.mapping._types import SceneLayer
from .._common import ServerConnection
from ._geodataservice import GeoData
from ._layerfactory import Service
from ..admin._services import Service as AdminService

def _str_replace(mystring, rd):
    """replaces a value based on a key/value pair where the
    key is the text to replace and the value is the new value.

    The find/replace is case insensitive.

    """
    import re
    patternDict = {}
    myDict = {}
    for key,value in rd.items():
        pattern = re.compile(re.escape(key), re.IGNORECASE)
        patternDict[value] = pattern
    for key in patternDict:
        regex_obj = patternDict[key]
        mystring = regex_obj.sub(key, mystring)
    return mystring

class AdminServiceFactory(type):
    """
    Generates an Administrative Service Object from a url or service object
    """
    def __call__(cls,
                 service,
                 gis,
                 initialize=False):
        """generates the proper type of layer from a given url"""

        url = service._url
        if isinstance(service, FeatureLayer) or \
           os.path.basename(url).isdigit():
            parent = Service(url=os.path.dirname(url), server=gis)
            return AdminServiceGen(parent, gis)
        elif isinstance(service, (NetworkDataset)):
            rd = {'naserver', 'MapServer'}
            url = _str_replace(url, rd)
            parent = Service(url=url, server=gis)
            return AdminServiceGen(parent, gis)
        else:
            rd = {'/rest/': '/admin/'}
            connection = service._con
            admin_url = "%s.%s" % (
                _str_replace(os.path.dirname(url), rd),
                os.path.basename(url))
            return AdminService(url=admin_url, gis=gis)
        return type.__call__(cls, service, gis, False)
###########################################################################
@add_metaclass(AdminServiceFactory)
class AdminServiceGen(object):
    """
    The Layer class allows users to pass a url, connection or other object
    to the class and get back properties and functions specifically related
    to the service.

    Inputs:
       url - internet address to the service
       server - Server class
       item - Portal or AGOL Item class
    """
    def __init__(self, service, gis):
        iterable = None
        if iterable is None:
            iterable = ()
        super(AdminService, self).__init__(service, gis)


import logging as _logging
import arcgis
from datetime import datetime
from arcgis. geoprocessing import import_toolbox
from arcgis.features import FeatureSet
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool

_log = _logging.getLogger(__name__)

_use_async = True

default_stops = {'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                                                {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString',
                                                 'length': 128}, {'alias': 'Route Name', 'name': 'RouteName',
                                                                  'type': 'esriFieldTypeString', 'length': 128},
                                                {'alias': 'Sequence', 'name': 'Sequence',
                                                 'type': 'esriFieldTypeInteger'},
                                                {'alias': 'Additional Time', 'name': 'AdditionalTime',
                                                 'type': 'esriFieldTypeDouble'},
                                                {'alias': 'Additional Distance', 'name': 'AdditionalDistance',
                                                 'type': 'esriFieldTypeDouble'},
                                                {'alias': 'Time Window Start', 'name': 'TimeWindowStart',
                                                 'type': 'esriFieldTypeDate', 'length': 8},
                                                {'alias': 'Time Window End', 'name': 'TimeWindowEnd',
                                                 'type': 'esriFieldTypeDate', 'length': 8},
                                                {'alias': 'Curb Approach', 'name': 'CurbApproach',
                                                 'type': 'esriFieldTypeSmallInteger'}],
                                     'geometryType': 'esriGeometryPoint', 'displayFieldName': '',
                                     'exceededTransferLimit': False,
                                     'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_point_barriers = {
                    'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                               {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString', 'length': 128},
                               {'alias': 'Barrier Type', 'name': 'BarrierType', 'type': 'esriFieldTypeInteger'},
                               {'alias': 'Additional Time', 'name': 'Additional_Time', 'type': 'esriFieldTypeDouble'},
                               {'alias': 'Additional Distance', 'name': 'Additional_Distance',
                                'type': 'esriFieldTypeDouble'},
                               {'alias': 'CurbApproach', 'name': 'CurbApproach', 'type': 'esriFieldTypeSmallInteger'}],
                    'geometryType': 'esriGeometryPoint', 'displayFieldName': '', 'exceededTransferLimit': False,
                    'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_line_barriers = {
                    'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                               {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString', 'length': 128},
                               {'alias': 'SHAPE_Length', 'name': 'SHAPE_Length', 'type': 'esriFieldTypeDouble'}],
                    'geometryType': 'esriGeometryPolyline', 'displayFieldName': '', 'exceededTransferLimit': False,
                    'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_polygon_barriers = {
                    'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                               {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString', 'length': 128},
                               {'alias': 'Barrier Type', 'name': 'BarrierType', 'type': 'esriFieldTypeInteger'},
                               {'alias': 'Scaled Time Factor', 'name': 'ScaledTimeFactor',
                                'type': 'esriFieldTypeDouble'},
                               {'alias': 'Scaled Distance Factor', 'name': 'ScaledDistanceFactor',
                                'type': 'esriFieldTypeDouble'},
                               {'alias': 'SHAPE_Length', 'name': 'SHAPE_Length', 'type': 'esriFieldTypeDouble'},
                               {'alias': 'SHAPE_Area', 'name': 'SHAPE_Area', 'type': 'esriFieldTypeDouble'}],
                    'geometryType': 'esriGeometryPolygon', 'displayFieldName': '', 'exceededTransferLimit': False,
                    'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_restrictions = """['Avoid Unpaved Roads', 'Avoid Private Roads', 'Driving an Automobile', 'Through Traffic Prohibited', 'Roads Under Construction Prohibited', 'Avoid Gates', 'Avoid Express Lanes', 'Avoid Carpool Roads']"""

default_attributes = {
                    'fields': [{'alias': 'ObjectID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                               {'alias': 'AttributeName', 'name': 'AttributeName', 'type': 'esriFieldTypeString',
                                'length': 255},
                               {'alias': 'ParameterName', 'name': 'ParameterName', 'type': 'esriFieldTypeString',
                                'length': 255},
                               {'alias': 'ParameterValue', 'name': 'ParameterValue', 'type': 'esriFieldTypeString',
                                'length': 25}], 'features': [{'attributes': {'OBJECTID': 1,
                                                                             'AttributeName': 'Any Hazmat Prohibited',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 2,
                                                                                'AttributeName': 'Avoid Carpool Roads',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 3,
                                                                             'AttributeName': 'Avoid Express Lanes',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 4,
                                                                                'AttributeName': 'Avoid Ferries',
                                                                                'ParameterValue': 'AVOID_MEDIUM',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 5,
                                                                             'AttributeName': 'Avoid Gates',
                                                                             'ParameterValue': 'AVOID_MEDIUM',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 6,
                                                                                'AttributeName': 'Avoid Limited Access Roads',
                                                                                'ParameterValue': 'AVOID_MEDIUM',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 7,
                                                                             'AttributeName': 'Avoid Private Roads',
                                                                             'ParameterValue': 'AVOID_MEDIUM',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 8,
                                                                                'AttributeName': 'Avoid Roads Unsuitable for Pedestrians',
                                                                                'ParameterValue': 'AVOID_HIGH',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 9,
                                                                             'AttributeName': 'Avoid Stairways',
                                                                             'ParameterValue': 'AVOID_HIGH',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 10,
                                                                                'AttributeName': 'Avoid Toll Roads',
                                                                                'ParameterValue': 'AVOID_MEDIUM',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 11,
                                                                             'AttributeName': 'Avoid Toll Roads for Trucks',
                                                                             'ParameterValue': 'AVOID_MEDIUM',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 12,
                                                                                'AttributeName': 'Avoid Truck Restricted Roads',
                                                                                'ParameterValue': 'AVOID_HIGH',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 13,
                                                                             'AttributeName': 'Avoid Unpaved Roads',
                                                                             'ParameterValue': 'AVOID_HIGH',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 14,
                                                                                'AttributeName': 'Axle Count Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Number of Axles'}}, {
                                                                 'attributes': {'OBJECTID': 15,
                                                                                'AttributeName': 'Axle Count Restriction',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 16,
                                                                             'AttributeName': 'Driving a Bus',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 17,
                                                                                'AttributeName': 'Driving a Delivery Vehicle',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 18,
                                                                             'AttributeName': 'Driving a Taxi',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 19,
                                                                                'AttributeName': 'Driving a Truck',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 20,
                                                                             'AttributeName': 'Driving an Automobile',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 21,
                                                                                'AttributeName': 'Driving an Emergency Vehicle',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 22,
                                                                             'AttributeName': 'Height Restriction',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 23,
                                                                                'AttributeName': 'Height Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Vehicle Height (meters)'}},
                                                             {'attributes': {'OBJECTID': 24,
                                                                             'AttributeName': 'Kingpin to Rear Axle Length Restriction',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 25,
                                                                                'AttributeName': 'Kingpin to Rear Axle Length Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Vehicle Kingpin to Rear Axle Length (meters)'}},
                                                             {'attributes': {'OBJECTID': 26,
                                                                             'AttributeName': 'Length Restriction',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 27,
                                                                                'AttributeName': 'Length Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Vehicle Length (meters)'}},
                                                             {'attributes': {'OBJECTID': 28,
                                                                             'AttributeName': 'Preferred for Pedestrians',
                                                                             'ParameterValue': 'PREFER_LOW',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 29,
                                                                                'AttributeName': 'Riding a Motorcycle',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 30,
                                                                             'AttributeName': 'Roads Under Construction Prohibited',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 31,
                                                                                'AttributeName': 'Semi or Tractor with One or More Trailers Prohibited',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 32,
                                                                             'AttributeName': 'Single Axle Vehicles Prohibited',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 33,
                                                                                'AttributeName': 'Tandem Axle Vehicles Prohibited',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 34,
                                                                             'AttributeName': 'Through Traffic Prohibited',
                                                                             'ParameterValue': 'AVOID_HIGH',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 35,
                                                                                'AttributeName': 'Truck with Trailers Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Number of Trailers on Truck'}},
                                                             {'attributes': {'OBJECTID': 36,
                                                                             'AttributeName': 'Truck with Trailers Restriction',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 37,
                                                                                'AttributeName': 'Use Preferred Hazmat Routes',
                                                                                'ParameterValue': 'PREFER_MEDIUM',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 38,
                                                                             'AttributeName': 'Use Preferred Truck Routes',
                                                                             'ParameterValue': 'PREFER_MEDIUM',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 39,
                                                                                'AttributeName': 'WalkTime',
                                                                                'ParameterValue': '5',
                                                                                'ParameterName': 'Walking Speed (km/h)'}},
                                                             {'attributes': {'OBJECTID': 40, 'AttributeName': 'Walking',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 41,
                                                                                'AttributeName': 'Weight Restriction',
                                                                                'ParameterValue': 'PROHIBITED',
                                                                                'ParameterName': 'Restriction Usage'}},
                                                             {'attributes': {'OBJECTID': 42,
                                                                             'AttributeName': 'Weight Restriction',
                                                                             'ParameterValue': '0',
                                                                             'ParameterName': 'Vehicle Weight (kilograms)'}},
                                                             {'attributes': {'OBJECTID': 43,
                                                                             'AttributeName': 'Weight per Axle Restriction',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 44,
                                                                                'AttributeName': 'Weight per Axle Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Vehicle Weight per Axle (kilograms)'}},
                                                             {'attributes': {'OBJECTID': 45,
                                                                             'AttributeName': 'Width Restriction',
                                                                             'ParameterValue': 'PROHIBITED',
                                                                             'ParameterName': 'Restriction Usage'}}, {
                                                                 'attributes': {'OBJECTID': 46,
                                                                                'AttributeName': 'Width Restriction',
                                                                                'ParameterValue': '0',
                                                                                'ParameterName': 'Vehicle Width (meters)'}}],
                    'displayFieldName': '', 'exceededTransferLimit': False}

default_tolerance = {'distance': 10, 'units': 'esriMeters'}





def find_routes(stops,
                measurement_units='Minutes',
                analysis_region=None,
                reorder_stops_to_find_optimal_routes=False,
                preserve_terminal_stops='Preserve First',
                return_to_start=False,
                use_time_windows=False,
                time_of_day=None,
                time_zone_for_time_of_day='Geographically Local',
                uturn_at_junctions='Allowed Only at Intersections and Dead Ends',
                point_barriers=None,
                line_barriers=None,
                polygon_barriers=None,
                use_hierarchy=True,
                restrictions=None,
                attribute_parameter_values=None,
                route_shape='True Shape',
                route_line_simplification_tolerance=None,
                populate_route_edges=False,
                populate_directions=True,
                directions_language='en',
                directions_distance_units='Miles',
                directions_style_name='NA Desktop',
                travel_mode='Custom',
                impedance='Drive Time',
                overrides=None,
                time_impedance='TravelTime',
                save_route_data=False,
                distance_impedance='Kilometers',
                output_format='Feature Set',
                save_output_na_layer=False,
                time_zone_for_time_windows='Geographically Local',
                gis=None,
                future=False):
    """
    Find Routes determines the shortest paths to visit the input stops and
    returns the driving directions, information about the visited stops,
    and the route paths, including travel time and distance., The tool is
    capable of finding routes that visit several input stops in a sequence
    you predetermine or in the  sequence that minimizes overall travel. You
    can group the input stops into different routes using the RouteName
    field, and the tool will output one route for each group of stops,
    allowing you to generate routes for many vehicles in a single solve
    operation.

    Parameters:

    stops: Stops (FeatureSet). Required parameter.  Specify the locations you want the output route
           or routes to visit. You can add up to 10,000 stops and assign up to 150 stops to a single
           route. (Assign stops to routes using the RouteName attribute.)When specifying the stops,
           you can set properties for each one, such as its name or service time, by using
           attributes. The stops can be specified with the following attributes:NameThe name of the
           stop. The name is used in the driving directions. If the name is not specified, a unique
           name prefixed with Location is automatically generated in the output stops, routes, and
           directions.RouteNameThe name of the route to which the stop is assigned. Assigning the
           same route name to different stops causes those stops to be grouped together and visited
           by the same route. You can generate many routes in a single solve by assigning unique
           route names to different groups of stops. You can group up to 150 stops into one route.
           SequenceThe output routes will visit the stops in the order you specify with this
           attribute. Within a group of stops that have the same RouteName value, the sequence
           number should be greater than 0 but not greater than the total number of stops. Also, the
           sequence number should not be duplicated. If Reorder Stops To Find Optimal Routes is
           checked (True), all but possibly the first and last sequence values for each route name
           are ignored so the tool can find the sequence that minimizes overall travel for each
           route. (The settings for Preserve Ordering of Stops and Return to Start determine whether
           the first or last sequence values for each route are ignored.)AdditionalTime The amount
           of time spent at the stop, which is added to the total time of the route. The default value
           is 0. The units for this attribute value are specified by the Measurement Units parameter.
           The attribute value is included in the analysis only when the measurement units are time
           based.  You can account for the extra time it takes at the stop to complete a task, such
           as to repair an appliance, deliver a package, or inspect the premises.
           AdditionalDistance, The extra distance traveled at the stops, which is added to the total
           distance of the route. The default value is 0.The units for this attribute value are
           specified by the Measurement Units parameter. The attribute value is included in the
           analysis only when the measurement units are distance based. Generally, the location of a
           stop, such as a home, isn't exactly on the street; it is set back somewhat from the road.
           This attribute value can be used to model the distance between the actual stop location
           and its location on the street if it is important to include that distance in the total
           travel distance.AdditionalCostThe extra cost spent at the stop, which is added to the
           total cost of the route. The default value is 0. This attribute value should be used when
           the travel mode for the analysis uses an impedance attribute that is neither time-based
           or distance-based The units for the attribute values are interpreted to be in unknown
           units. TimeWindowStartThe earliest time the stop can be visited. By specifying a start
           and end time for a stop's time window, you are defining when a route should visit the
           stop. When the travel mode for the analysis uses an impedance attribute that is
           time-based, by specifying time-window values the analysis will find a solution that
           minimizes overall travel and reaches the stop within the prescribed time window.  Make
           sure you specify the value as a date and time value, such as 8/12/2015 12:15 PM.  When
           solving a problem that spans multiple time zones, time-window values refer to the time
           zone in which the stop is located.This field can contain a null value; a null value
           indicates a route can arrive at any time before the time indicated in the TimeWindowEnd
           attribute. If a null value is also present in TimeWindowEnd, a route can visit the stop
           at any time.TimeWindowEndThe latest time the stop can be visited. By specifying a start
           and end time for a stop's time window, you are defining when a route should visit the
           stop. When the travel mode for the analysis uses an impedance attribute that is
           time-based, specifying time-window values will cause the analysis to find a solution
           that minimizes overall travel and reaches the stop within the prescribed time window.
           Make sure you specify the value as a date and time value, such as 8/12/2015 12:15 PM.
           When solving a problem that spans multiple time zones, time-window values refer to the
           time zone in which the stop is located.This field can contain a null value; a null value
           indicates a route can arrive at any time after the time indicated in the TimeWindowStart
           attribute. If a null value is also present in TimeWindowStart, a route can visit the
           stop at any time.CurbApproachSpecifies the direction a vehicle may arrive at and depart
           from the stop. The field value is specified as one of the
           following integers (use the numeric code, not the name in parentheses):  0 (Either side of vehicle) - The vehicle can approach and depart the stop in either direction, so a U-turn is allowed at the stop. This setting can be chosen if it is possible and practical for a vehicle to turn around at the stop. This decision may depend on the width of the road and the amount of traffic or whether the stop has a parking lot where vehicles can enter and turn around. 1 (Right side of vehicle)-When the vehicle approaches and departs the stop, the curb must be on the right side of the vehicle. A U-turn is prohibited. This is typically used for vehicles such as buses that must arrive with the bus stop on the right-hand side. 2 (Left side of vehicle)-When the vehicle approaches and departs
           the stop, the curb must be on the left side of the vehicle. A
           U-turn is prohibited. This is typically used for vehicles such as buses that must arrive with the bus stop on the left-hand side.  3 (No U-Turn)-When
           the vehicle approaches the stop, the curb can be on either side
           of the vehicle; however, the vehicle must depart without turning
           around. The CurbApproach property is designed to work with both kinds of national driving standards: right-hand traffic (United States) and left-hand traffic (United Kingdom). First, consider a stop on the left side of a vehicle. It is always on the left side regardless of whether the vehicle travels on the left or right half of the road. What may change with national driving standards is your decision to approach a stop from one of two directions, that is, so it ends up on the right or left side of the vehicle. For example, if you want to arrive at a stop and not have a lane of traffic between the vehicle and the stop, you would choose Right side of vehicle (1) in the United States but Left side of vehicle (2) in the United Kingdom.LocationType Specifies the stop type. The field value is specified as one of the following integers (use the numeric code, not the name in the parentheses): 0 (Stop)-A location that the route should visit. This is the default.1 (Waypoint)- A location that the route should travel through without making a stop. Waypoints can be used to force the route to take a specific path (to go through the waypoint) without being considered an actual stop. Waypoints do not appear in directions.BearingThe direction in which a point is moving. The units are degrees and are measured clockwise from true north. This field is used in conjunction with the BearingTol field. Bearing data is usually sent automatically from a mobile device equipped with a GPS receiver. Try to include bearing data if you are loading an input location that is moving, such as a pedestrian or a vehicle. Using this field tends to prevent adding locations to the wrong edges, which can occur when a vehicle is near an intersection or an overpass for example. Bearing also helps the tool determine on which side of the street the point is. For more information, see Bearing and BearingTol in the ArcGIS help system. BearingTolThe bearing tolerance value creates a range of acceptable bearing values when locating moving points on an edge using the Bearing field. If the value from the Bearing field is within the range of acceptable values that are generated from the bearing tolerance on an edge, the point can be added as a network location there; otherwise, the closest point on the next-nearest edge is evaluated. The units are in degrees and the default value is 30. Values must be greater than zero and less than 180. A value of 30 means that when ArcGIS Network Analyst extension  attempts to add a network location on an edge, a range of acceptable bearing values is generated 15 degrees to either side of the edge (left and right) and in both digitized directions of the edge. For more information, see Bearing and BearingTol in the ArcGIS help system. NavLatencyThis field is only used in the solve process if Bearing and BearingTol also have values; however, entering a NavLatency value is optional, even when values are present in Bearing and BearingTol. NavLatency indicates how much time is expected to elapse from the moment GPS information is sent from a moving vehicle to a server and the moment the processed route is received by the vehicle's navigation device. The time units of NavLatency are the same as the units specified by the timeUnits property of the analysis object.

           measurement_units: Measurement Units (str). Required parameter.  Specify the units that should be used to measure and report the total travel time or travel distance for the output routes. The units you choose for this parameter determine whether the tool will measure distance or time to find the best routes. Choose a time unit to minimize travel time for your chosen travel mode (driving or walking time, for instance). To minimize travel distance for the given travel mode, choose a distance unit. Your choice also determines in which units the tool will report total time or distance in the results. The choices include the following:MetersKilometersFeetYardsMilesNauticalMilesSecondsMinutesHoursDays
                              Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'NauticalMiles', 'Seconds', 'Minutes', 'Hours', 'Days']

           analysis_region: Analysis Region (str). Optional parameter.  Specify the region in which to perform the analysis. If a value is not specified for this parameter, the tool
                            will automatically calculate the region name based on the location of the input points. Setting the name of the region is required only if the
                            auto-detection of the region name is not accurate for your inputs. To specify a region, use one of the following values:  Europe,Japan,Korea,MiddleEast,
                            and Africa.  NorthAmerica, SouthAmerica, SouthAsia,Thailand; The following region names are no longer supported and will be removed in future releases.
                            If you specify one of the deprecated region names, the tool automatically assigns a supported region name for your region. Greece redirects to Europe.
                            India redirects to South Asia. Oceania redirects to South Asia. South East Asia redirects to South Asia. Taiwan redirects to South Asia.

                            Choice list:['Europe', 'Japan', 'Korea', 'MiddleEastAndAfrica', 'NorthAmerica', 'SouthAmerica', 'SouthAsia', 'Thailand']

           reorder_stops_to_find_optimal_routes: Reorder Stops to Find Optimal Routes (bool). Optional parameter.  Specify whether to visit the stops in the order you define or the order the tool determines will minimize overall travel. Checked (True)-The tool determines the sequence that will minimize overall travel distance or time. It can reorder stops and account for time windows at stops. Additional parameters allow you to preserve the first or last stops while allowing the tool to reorder the intermediary stops.Unchecked (False)-The stops are visited in the order you define. This is the default option. You can set the order of stops using a Sequence attribute in the input stops features or let the sequence be determined by the Object ID of the stops. Finding the optimal stop order and the best routes is commonly known as solving the traveling salesman problem (TSP).

           preserve_terminal_stops: Preserve Terminal Stops (str). Optional parameter.  When Reorder Stops to Find Optimal Routes is checked (or True), you have options to preserve the starting or ending stops and the tool can reorder the rest. The first and last stops are determined by their Sequence attribute values or, if the Sequence values are null, by their Object ID values.Preserve First-The tool won't reorder the first stop. Choose this option if you are starting from a known location, such as your home, headquarters, or current location.Preserve Last-The tool won't reorder the last stop. The output routes may start from any stop feature but must end at the predetermined last stop. Preserve First and Last-The tool won't reorder the first and last stops.Preserve None-The tool may reorder any stop, including the first and last stops. The route may start or end at any of the stop features.Preserve Terminal Stops is ignored when Reorder Stops to Find Optimal Routes is unchecked (or False).
                  Choice list:['Preserve First', 'Preserve Last', 'Preserve First and Last', 'Preserve None']

           return_to_start: Return to Start (bool). Optional parameter.  Choose whether routes should start and end at the same location. With this option you can avoid duplicating the first stop feature and sequencing the duplicate stop at the end.The starting location of the route is the stop feature with the lowest value in the Sequence attribute. If the Sequence values are null, it is the stop feature with the lowest Object ID value.Checked (True)-The route should start and end at the first stop feature. This is the default value. When Reorder Stops to Find Optimal Routes and Return to Start are both checked (or True), Preserve Terminal Stops must be set to Preserve First.Unchecked (False)-The route won't start and end at the first stop feature.

           use_time_windows: Use Time Windows (bool). Optional parameter.  Check this option (or set it to True) if any input stops have time windows that specify when the route should reach the stop. You can add time windows to input stops by entering time values in the TimeWindowStart and TimeWindowEnd attributes.Checked (True)-The input stops have time windows and you want the tool to try to honor them.Unchecked (False)-The input stops don't have time windows, or if they do, you don't want the tool to try to honor them. This is the default value.The tool will take slightly longer to run when Use Time Windows is checked (or True), even when none of the input stops have time windows, so it is recommended to uncheck this option (set to False) if possible.

           time_of_day: Time of Day (datetime). Optional parameter.  Specifies the time and date at which the routes should
            begin. If you are modeling the driving travel mode and specify the current date and time as the value
            for this parameter, the tool will use live traffic conditions to
            find the best routes and the total travel time will be based
            on traffic conditions. Specifying a time of day results in more accurate
            routes and estimations of travel times because the
            travel times account for the traffic conditions that are applicable
            for that date and time.The Time Zone for Time of Day parameter specifies whether this time and date refer to UTC or the time zone in which the stop is located.The tool ignores this parameter when Measurement Units isn't set to a time-based unit.

           time_zone_for_time_of_day: Time Zone for Time of Day (str). Optional parameter.  Specifies the time zone of the Time of Day parameter.Geographically Local-The Time of Day parameter refers to the time zone in which the first stop of a route is located. If you are generating many routes that start in multiple times zones, the start times are staggered in Coordinated Universal Time (UTC). For example, a Time of Day value of 10:00 a.m., 2 January, would mean a start time of 10:00 a.m. Eastern Standard Time (UTC-3:00) for routes beginning in the Eastern Time Zone and 10:00 a.m. Central Standard Time (UTC-4:00) for routes beginning in the Central Time Zone. The start times are offset by one hour in UTC. The arrive and depart times and dates recorded in the output Stops feature class will refer to the local time zone of the first stop for each route.UTC-The Time of Day parameter refers to Coordinated Universal Time (UTC). Choose this option if you want to generate a route for a specific time, such as now, but aren't certain in which time zone the first stop will be located. If you are generating many routes spanning multiple times zones, the start times in UTC are simultaneous. For example, a Time of Day value of 10:00 a.m., 2 January, would mean a start time of 5:00 a.m. Eastern Standard Time (UTC-5:00) for routes beginning in the Eastern Time Zone and 4:00 a.m. Central Standard Time (UTC-6:00) for routes beginning in the Central Time Zone. Both routes would start at 10:00 a.m. UTC. The arrive and depart times and dates recorded in the output Stops feature class will refer to UTC.
                  Choice list:['Geographically Local', 'UTC']

           uturn_at_junctions: UTurn at Junctions (str). Optional parameter.  The U-Turn policy at junctions. Allowing U-turns implies the solver can turn around at a junction and double back on the same street.

           Given that junctions represent street intersections and dead ends, different vehicles  may be able to turn around at some junctions but not at others-it depends on whether the junction represents an intersection or dead end. To accommodate, the U-turn policy parameter is implicitly specified by how many edges connect to the junction, which is known as junction valency. The acceptable values for this parameter are listed below; each is followed by a description of its meaning in terms of junction valency.
            Allowed- U-turns are permitted at junctions with any number of connected edges. This is the default value.Not Allowed- U-turns are prohibited at all junctions, regardless of junction valency. Note, however, that U-turns are still permitted at network locations even when this setting is chosen; however, you can set the individual network locations' CurbApproach property to prohibit U-turns there as well.Allowed only at Dead Ends- U-turns are prohibited at all junctions, except those that have only one adjacent edge (a dead end).Allowed only at Intersections and Dead Ends- U-turns are prohibited at junctions where exactly two adjacent edges meet but are permitted at intersections (junctions with three or more adjacent edges) and dead ends (junctions with exactly one adjacent edge).  Often, networks have extraneous junctions in the middle of road segments. This option prevents vehicles from making U-turns at these locations.
                  Choice list:['Allowed', 'Not Allowed', 'Allowed Only at Dead Ends', 'Allowed Only at Intersections and Dead Ends']

           point_barriers: Point Barriers (FeatureSet). Optional parameter.  Specify one or more points to act as temporary
            restrictions or represent additional time or distance that may be
            required to travel on the underlying streets. For example, a point
            barrier can be used to represent a fallen tree along a street or
            time delay spent at a railroad crossing. The tool imposes a limit of 250 points that can be added
            as barriers.When specifying the point barriers, you can set properties for each one, such as its name or barrier type, by using attributes. The point barriers can be specified with the following attributes:
                                                                    Name The name of the barrier. BarrierType Specifies whether the point barrier restricts travel
            completely or adds time or distance when it is crossed. The value
            for this attribute is specified as one of the following
            integers (use the numeric code, not the name in parentheses):  0 (Restriction)-Prohibits travel through the barrier. The barrier
            is referred to as a restriction point barrier since it acts as a
            restriction.  2 (Added Cost)-Traveling through the barrier increases the travel
            time or distance by the amount specified in the
            Additional_Time, Additional_Distance, or Additional_Cost field. This barrier type is
            referred to as an added-cost point barrier.  Additional_Time Indicates how much travel time is added when the
            barrier is traversed. This field is applicable only for added-cost
            barriers and only if the travel mode used for the analysis uses an impedance attribute that is time-based. This field
            value must be greater than or equal to zero, and its units are the same as those specified in the
            Measurement Units parameter. Additional_DistanceIndicates how much distance is added when the
            barrier is traversed. This field is applicable only for added-cost
            barriers and only if the travel mode used for the analysis uses an impedance attribute that is distance-based. The field value
            must be greater than or equal to zero, and its units are the same as those specified in the
            Measurement Units parameter.Additional_CostIndicates how much cost is added when the
            barrier is traversed. This field is applicable only for added-cost
            barriers and only if the travel mode used for the analysis uses an impedance attribute that is neither time-based or distance-based. FullEdgeSpecify how the restriction point barriers are applied to the edge elements during the analysis. The field value is specified as one of the following integers (use the numeric code, not the name in parentheses):  0 (False):  Permits travel on the edge up to the barrier, but not through it. This is the default value.1 (True): Restricts travel anywhere on the associated edge. CurbApproachSpecifies the direction of traffic that is affected by the barrier. The field value is specified as one of the following integers (use the numeric code, not the name in parentheses): 0 (Either side of vehicle): The barrier affects travel over the edge in both directions.1 (Right side of vehicle): Vehicles are only affected if the barrier is on their right side during the approach. Vehicles that traverse the same edge but approach the barrier on their left side are not affected by the barrier. 2 (Left side of vehicle): Vehicles are only affected if the barrier is on their left side during the approach. Vehicles that traverse the same edge but approach the barrier on their right side are not affected by the barrier. Since junctions are points and don't have a side, barriers on junctions affect all vehicles regardless of the curb approach. The CurbApproach property was designed to work with both kinds of national driving standards: right-hand traffic (United States) and left-hand traffic (United Kingdom). First, consider a facility on the left side of a vehicle. It is always on the left side regardless of whether the vehicle travels on the left or right half of the road. What may change with national driving standards is your decision to approach a facility from one of two directions, that is, so it ends up on the right or left side of the vehicle. For example, if you want to arrive at a facility and not have a lane of traffic between the vehicle and the facility, you would choose Right side of vehicle (1) in the United States but Left side of vehicle (2) in the United Kingdom.BearingThe direction in which a point is moving. The units are degrees and are measured clockwise from true north. This field is used in conjunction with the BearingTol field. Bearing data is usually sent automatically from a mobile device equipped with a GPS receiver. Try to include bearing data if you are loading an input location that is moving, such as a pedestrian or a vehicle. Using this field tends to prevent adding locations to the wrong edges, which can occur when a vehicle is near an intersection or an overpass for example. Bearing also helps the tool determine on which side of the street the point is. For more information, see Bearing and BearingTol in the ArcGIS help system. BearingTolThe bearing tolerance value creates a range of acceptable bearing values when locating moving points on an edge using the Bearing field. If the value from the Bearing field is within the range of acceptable values that are generated from the bearing tolerance on an edge, the point can be added as a network location there; otherwise, the closest point on the next-nearest edge is evaluated. The units are in degrees and the default value is 30. Values must be greater than zero and less than 180. A value of 30 means that when ArcGIS Network Analyst extension  attempts to add a network location on an edge, a range of acceptable bearing values is generated 15 degrees to either side of the edge (left and right) and in both digitized directions of the edge. For more information, see Bearing and BearingTol in the ArcGIS help system. NavLatencyThis field is only used in the solve process if Bearing and BearingTol also have values; however, entering a NavLatency value is optional, even when values are present in Bearing and BearingTol. NavLatency indicates how much time is expected to elapse from the moment GPS information is sent from a moving vehicle to a server and the moment the processed route is received by the vehicle's navigation device. The time units of NavLatency are the same as the units specified by the timeUnits property of the analysis object.

           line_barriers: Line Barriers (FeatureSet). Optional parameter.  Specify one or more lines that prohibit travel anywhere
            the lines intersect the streets. For example, a parade or protest
            that blocks traffic across several street segments can be modeled
            with a line barrier. A line barrier can also quickly fence off
            several roads from being traversed, thereby channeling possible
            routes away from undesirable parts of the street
            network. The tool imposes a limit on the number of streets you can
            restrict using the Line Barriers parameter. While there is no limit on
            the number of lines you can specify as line barriers, the combined
            number of streets intersected by all the lines cannot exceed
            500.When specifying the line barriers, you can set a name property for each one by using the following attribute:
                                                                    Name The name of the barrier.

           polygon_barriers: Polygon Barriers (FeatureSet). Optional parameter.  Specify polygons that either completely restrict travel or
            proportionately scale the time or distance required to travel on
            the streets intersected by the polygons. The service imposes a limit on the number of streets you
            can restrict using the Polygon Barriers parameter. While there is
            no limit on the number of polygons you can specify as the polygon
            barriers, the combined number of streets intersected by all the
            polygons should not exceed 2,000.When specifying the polygon barriers, you can set properties for each one, such as its name or barrier type, by using attributes. The polygon barriers can be specified with the following attributes:
            Name The name of the barrier. BarrierType Specifies whether the barrier restricts travel completely
            or scales the time or distance for traveling through it. The field
            value is specified as one of the following integers (use the numeric code, not the name in parentheses):  0 (Restriction)-Prohibits traveling through any part of the barrier.
            The barrier is referred to as a restriction polygon barrier since it
            prohibits traveling on streets intersected by the barrier. One use
            of this type of barrier is to model floods covering areas of the
            street that make traveling on those streets impossible.  1 (Scaled Cost)-Scales the time or distance required to travel the
            underlying streets by a factor specified using the ScaledTimeFactor
            or ScaledDistanceFactor field. If the streets are partially
            covered by the barrier, the travel time or distance is apportioned
            and then scaled. For example, a factor 0.25 would mean that travel
            on underlying streets is expected to be four times faster than
            normal. A factor of 3.0 would mean it is expected to take three
            times longer than normal to travel on underlying streets. This
            barrier type is referred to as a scaled-cost polygon barrier. It
            might be used to model storms that reduce travel speeds in specific
            regions. ScaledTimeFactor This is the factor by which the travel time of the streets
            intersected by the barrier is multiplied. The field value must be greater than zero. This field is applicable only for scaled-cost
            barriers and only if the travel mode used for the analysis uses an impedance attribute that is time based. ScaledDistanceFactor This is the factor by which the distance of the streets
            intersected by the barrier is multiplied. The field value must be greater than zero.This field is applicable only for scaled-cost
            barriers and only if the travel mode used for the analysis uses an impedance attribute that is distance based.  ScaledCostFactor This is the factor by which the cost of the streets
            intersected by the barrier is multiplied. The field value must be greater than zero. This field is applicable only for scaled-cost
            barriers and only if the travel mode used for the analysis uses an impedance attribute that is neither time based nor distance based.

           use_hierarchy: Use Hierarchy (bool). Optional parameter.  Specify whether hierarchy should be used when finding the shortest paths between stops. Checked (True)-Use hierarchy when finding routes. When
            hierarchy is used, the tool prefers higher-order streets (such as
            freeways) to lower-order streets (such as local roads), and can be used
            to simulate the driver preference of traveling on freeways instead
            of local roads even if that means a longer trip. This is especially
            true when finding routes to faraway locations, because drivers on long-distance trips tend to prefer traveling on freeways where stops, intersections, and turns can be avoided. Using hierarchy is computationally faster,
            especially for long-distance routes, since the tool can determine the
            best route from a relatively smaller subset of streets.   Unchecked (False)-Do not use hierarchy when finding routes. If
            hierarchy is not used, the tool considers all the streets and doesn't
            prefer higher-order streets when finding the route. This is often
            used when finding short routes within a city.  The tool automatically reverts to using hierarchy if the
            straight-line distance between facilities and demand points is
            greater than 50 miles (80.46
            kilometers), even if you have set this parameter to not use hierarchy.This parameter is ignored unless Travel Mode is set to Custom. When modeling a custom walking mode, it is recommended to turn off hierarchy since the hierarchy is designed for motorized vehicles.

           restrictions: Restrictions (str). Optional parameter.  Specify which restrictions should be honored by the tool when finding the best routes.A restriction represents a driving
            preference or requirement. In most cases, restrictions cause roads
            to be prohibited. For instance, using an Avoid Toll Roads restriction will result in a route that will include toll roads only when it is absolutely required to travel on toll roads in order to visit an incident or a facility. Height Restriction makes it possible to route around any clearances that are lower than the height of your vehicle. If you are carrying corrosive materials on your vehicle, using the Any Hazmat Prohibited restriction prevents hauling the materials along roads where it is marked as illegal to do so. The values you provide for this parameter are ignored unless Travel Mode is set to Custom.Some restrictions require an additional value to be
            specified for their desired use. This value needs to be associated
            with the restriction name and a specific parameter intended to work
            with the restriction. You can identify such restrictions if their
            names appear under the AttributeName column in the Attribute
            Parameter Values parameter. The ParameterValue field should be
            specified in the Attribute Parameter Values parameter for the
            restriction to be correctly used when finding traversable roads.Some restrictions are supported only in certain countries; their availability is stated by region in the list below. Of the restrictions that have limited availability within a region, you can check whether the restriction is available in a particular country by looking at the table in the Country List section of the Data coverage for network analysis services web page. If a country has a value of  Yes in the Logistics Attribute column, the restriction with select availability in the region is supported in that country. If you specify restriction names that are not available in the country where your incidents are located, the service ignores the invalid restrictions. The service also ignores restrictions whose Restriction Usage parameter value is between 0 and 1 (see the Attribute Parameter Value parameter). It prohibits all restrictions whose Restriction Usage parameter value is greater than 0.The tool supports the following restrictions: Any Hazmat Prohibited-The results will not include roads
            where transporting any kind of hazardous material is
            prohibited. Availability: Select countries in North America and Europe Avoid Carpool Roads-The results will avoid roads that are
            designated exclusively for carpool (high-occupancy)
            vehicles. Availability: All countries Avoid Express Lanes-The results will avoid roads designated
            as express lanes. Availability: All countries Avoid Ferries-The results will avoid ferries. Availability: All countries Avoid Gates-The results will avoid roads where there are
            gates such as keyed access or guard-controlled
            entryways.Availability: All countries Avoid Limited Access Roads-The results will avoid roads
            that are limited access highways. Availability: All countries Avoid Private Roads-The results will avoid roads that are
            not publicly owned and maintained. Availability: All countries Avoid Roads Unsuitable for Pedestrians-The results will avoid roads that are
            unsuitable for pedestrians. Availability: All countries Avoid Stairways-The results will avoid all stairways on a pedestrian suitable route. Availability: All countries Avoid Toll Roads-The results will avoid all toll
            roads for automobiles.Availability: All countries Avoid Toll Roads for Trucks-The results will avoid all toll
            roads for trucks.Availability: All countries Avoid Truck Restricted Roads-The results will avoid roads where trucks are not allowed, except when making deliveries.Availability: All countries Avoid Unpaved Roads-The results will avoid roads that are
            not paved (for example, dirt, gravel, and so on). Availability: All countries Axle Count Restriction-The results will not include roads
            where trucks with the specified number of axles are prohibited. The
            number of axles can be specified using the Number of Axles
            restriction parameter.Availability: Select countries in North America and Europe Driving a Bus-The results will not include roads where
            buses are prohibited. Using this restriction will also ensure that
            the results will honor one-way streets. Availability: All countries Driving a Delivery Vehicle-The results will not include
            roads where delivery vehicles are prohibited. Using this restriction
            will also ensure that the results will honor one-way
            streets. Availability: All countries Driving a Taxi-The results will not include roads where
            taxis are prohibited. Using this restriction will also ensure that
            the results will honor one-way streets. Availability: All countries Driving a Truck-The results will not include roads where
            trucks are prohibited. Using this restriction will also ensure that
            the results will honor one-way streets. Availability: All countries Driving an Automobile-The results will not include roads
            where automobiles are prohibited. Using this restriction will also
            ensure that the results will honor one-way streets. Availability: All countries Driving an Emergency Vehicle-The results will not include
            roads where emergency vehicles are prohibited. Using this
            restriction will also ensure that the results will honor one-way
            streets. Availability: All countries Height Restriction-The results will not include roads
            where the vehicle height exceeds the maximum allowed height for the
            road. The vehicle height can be specified using the Vehicle Height
            (meters) restriction parameter. Availability: Select countries in North America and Europe Kingpin to Rear Axle Length Restriction-The results will
            not include roads where the vehicle length exceeds the maximum
            allowed kingpin to rear axle for all trucks on the road. The length
            between the vehicle kingpin and the rear axle can be specified
            using the Vehicle Kingpin to Rear Axle Length (meters) restriction
            parameter. Availability: Select countries in North America and Europe Length Restriction-The results will not include roads
            where the vehicle length exceeds the maximum allowed length for the
            road. The vehicle length can be specified using the Vehicle Length
            (meters) restriction parameter. Availability: Select countries in North America and Europe Preferred for Pedestrians-The results will use preferred routes suitable for pedestrian navigation. Availability: Select countries in North America and Europe Riding a Motorcycle-The results will not include roads
            where motorcycles are prohibited. Using this restriction will also
            ensure that the results will honor one-way streets. Availability: All countries Roads Under Construction Prohibited-The results will not
            include roads that are under construction.Availability: All countries Semi or Tractor with One or More Trailers Prohibited-The
            results will not include roads where semis or tractors with one or
            more trailers are prohibited. Availability: Select countries in North America and Europe Single Axle Vehicles Prohibited-The results will not
            include roads where vehicles with single axles are
            prohibited.Availability: Select countries in North America and Europe Tandem Axle Vehicles Prohibited-The results will not
            include roads where vehicles with tandem axles are
            prohibited.Availability: Select countries in North America and Europe Through Traffic Prohibited-The results will not include
            roads where through traffic (non local) is prohibited.Availability: All countries Truck with Trailers Restriction-The results will not
            include roads where trucks with the specified number of trailers on
            the truck are prohibited. The number of trailers on the truck can
            be specified using the Number of Trailers on Truck restriction
            parameter.Availability: Select countries in North America and Europe Use Preferred Hazmat Routes-The results will prefer roads
            that are designated for transporting any kind of hazardous
            materials. Availability: Select countries in North America and Europe Use Preferred Truck Routes-The results will prefer roads
            that are designated as truck routes, such as the roads that are
            part of the national network as specified by the National Surface
            Transportation Assistance Act in the United States, or roads that
            are designated as truck routes by the state or province, or roads
            that are preferred by the trucks when driving in an
            area.Availability: Select countries in North America and Europe Walking-The results will not include roads where
            pedestrians are prohibited.Availability: All countries Weight Restriction-The results will not include roads
            where the vehicle weight exceeds the maximum allowed weight for the
            road. The vehicle weight can be specified using the Vehicle Weight
            (kilograms) restriction parameter.Availability: Select countries in North America and Europe Weight per Axle Restriction-The results will not include
            roads where the vehicle weight per axle exceeds the maximum allowed
            weight per axle for the road. The vehicle weight per axle can be
            specified using the Vehicle Weight per Axle (kilograms) restriction
            parameter.Availability: Select countries in North America and Europe Width Restriction-The results will not include roads where
            the vehicle width exceeds the maximum allowed width for the road.
            The vehicle width can be specified using the Vehicle Width (meters)
            restriction parameter.Availability: Select countries in North America and Europe The Driving a Delivery Vehicle restriction attribute is no longer available. The service will ignore this restriction since it is invalid. To achieve similar results, use the Driving a Truck restriction attribute along with the Avoid Truck Restricted Roads restriction attribute.
                  Choice list:['Any Hazmat Prohibited', 'Avoid Carpool Roads', 'Avoid Express Lanes', 'Avoid Ferries', 'Avoid Gates', 'Avoid Limited Access Roads', 'Avoid Private Roads', 'Avoid Roads Unsuitable for Pedestrians', 'Avoid Stairways', 'Avoid Toll Roads', 'Avoid Toll Roads for Trucks', 'Avoid Truck Restricted Roads', 'Avoid Unpaved Roads', 'Axle Count Restriction', 'Driving a Bus', 'Driving a Taxi', 'Driving a Truck', 'Driving an Automobile', 'Driving an Emergency Vehicle', 'Height Restriction', 'Kingpin to Rear Axle Length Restriction', 'Length Restriction', 'Preferred for Pedestrians', 'Riding a Motorcycle', 'Roads Under Construction Prohibited', 'Semi or Tractor with One or More Trailers Prohibited', 'Single Axle Vehicles Prohibited', 'Tandem Axle Vehicles Prohibited', 'Through Traffic Prohibited', 'Truck with Trailers Restriction', 'Use Preferred Hazmat Routes', 'Use Preferred Truck Routes', 'Walking', 'Weight Restriction', 'Weight per Axle Restriction', 'Width Restriction']

           attribute_parameter_values: Attribute Parameter Values (FeatureSet). Optional parameter.  Specify additional values required by some restrictions, such as the weight of a vehicle for Weight Restriction. You can also use the attribute parameter to specify whether any restriction prohibits, avoids, or prefers
            travel on roads that use the restriction. If the restriction is
            meant to avoid or prefer roads, you can further specify the degree
            to which they are avoided or preferred using this
            parameter. For example, you can choose to never use toll roads, avoid them as much as possible, or even highly prefer them.The values you provide for this parameter are ignored unless Travel Mode is set to Custom. If you specify the Attribute Parameter Values parameter from a
            feature class, the field names on the feature class must match the fields as described below:AttributeName- Lists the name of the restriction.ParameterName- Lists the name of the parameter associated with the
            restriction. A restriction can have one or more ParameterName field
            values based on its intended use.ParameterValue- The value for ParameterName used by the tool
            when evaluating the restriction. Attribute Parameter Values is dependent on the
            Restrictions parameter. The ParameterValue field is applicable only
            if the restriction name is specified as the value for the
            Restrictions parameter. In Attribute Parameter Values, each
            restriction (listed as AttributeName) has a ParameterName field
            value, Restriction Usage, that specifies whether the restriction
            prohibits, avoids, or prefers travel on the roads associated with
            the restriction and the degree to which the roads are avoided or
            preferred. The Restriction Usage ParameterName can be assigned any of
            the following string values or their equivalent numeric values
            listed within the parentheses:  PROHIBITED (-1)-Travel on the roads using the restriction is completely
            prohibited.  AVOID_HIGH (5)-It
            is highly unlikely for the tool to include in the route the roads
            that are associated with the restriction.  AVOID_MEDIUM (2)-It
            is unlikely for the tool to include in the route the roads that are
            associated with the restriction.  AVOID_LOW (1.3)-It
            is somewhat unlikely for the tool to include in the route the roads
            that are associated with the restriction.  PREFER_LOW (0.8)-It
            is somewhat likely for the tool to include in the route the roads
            that are associated with the restriction.  PREFER_MEDIUM (0.5)-It is likely for the tool to include in the route the roads that
            are associated with the restriction.  PREFER_HIGH (0.2)-It is highly likely for the tool to include in the route the roads
            that are associated with the restriction.  In most cases, you can use the default value, PROHIBITED,
            for the Restriction Usage if the restriction is dependent on a
            vehicle-characteristic such as vehicle height. However, in some
            cases, the value for Restriction Usage depends on your routing
            preferences. For example, the Avoid Toll Roads restriction has the
            default value of AVOID_MEDIUM for the Restriction Usage parameter.
            This means that when the restriction is used, the tool will try to
            route around toll roads when it can. AVOID_MEDIUM also indicates
            how important it is to avoid toll roads when finding the best
            route; it has a medium priority. Choosing AVOID_LOW would put lower
            importance on avoiding tolls; choosing AVOID_HIGH instead would
            give it a higher importance and thus make it more acceptable for
            the service to generate longer routes to avoid tolls. Choosing
            PROHIBITED would entirely disallow travel on toll roads, making it
            impossible for a route to travel on any portion of a toll road.
            Keep in mind that avoiding or prohibiting toll roads, and thus
            avoiding toll payments, is the objective for some; in contrast,
            others prefer to drive on toll roads because avoiding traffic is
            more valuable to them than the money spent on tolls. In the latter
            case, you would choose PREFER_LOW, PREFER_MEDIUM, or PREFER_HIGH as
            the value for Restriction Usage. The higher the preference, the
            farther the tool will go out of its way to travel on the roads
            associated with the restriction.

           route_shape: Route Shape (str). Optional parameter.  Specify the type of route features that are output by the
            tool. The parameter can be specified using one of the following
            values:  True Shape-Return the exact shape of the resulting route
            that is based on the underlying streets.  True Shape with Measures-Return the exact shape of the resulting route that is based on the underlying streets. Additionally, construct measures so the shape may be used in linear referencing. The measurements increase from the first stop and record the cumulative travel time or travel distance in the units specified by the Measurement Units parameter. Straight Line-Return a straight line between two stops.   None-Do not return any shapes for the routes. This value
            can be useful, and return results quickly, in cases where you are only interested in determining
            the total travel time or travel distance of a route.  When the Route Shape parameter is set to True Shape or True Shape with Measures, the
            generalization of the route shape can be further controlled using
            the appropriate value for the Route Line Simplification Tolerance
            parameter. No matter which value you choose for the Route Shape
            parameter, the best route is always determined by minimizing the
            travel time or the travel distance, never using the straight-line
            distance between stops. This means that only the route shapes are different,
            not the underlying streets that are searched when finding the
            route.
                  Choice list:['True Shape', 'True Shape with Measures', 'Straight Line', 'None']

           route_line_simplification_tolerance: Route Line Simplification Tolerance (LinearUnit). Optional parameter.  Specify by how much you want to simplify the geometry of the output lines for routes, directions, and route edges.The tool ignores this parameter if the Route Shape parameter isn't set to True Shape. Simplification maintains critical
            points on a route, such as turns at intersections, to define the
            essential shape of the route and removes other points. The
            simplification distance you specify is the maximum allowable offset
            that the simplified line can deviate from the original line.
            Simplifying a line reduces the number of vertices that are part of
            the route geometry. This improves the tool execution
            time.

           populate_route_edges: Populate Route Edges (bool). Optional parameter.  Specify whether the tool should generate edges for each route. Route edges represent the individual street features or other similar features that are traversed by a route. The output Route Edges layer is commonly used to see which streets or paths are traveled on the most or least by the resultant routes. Checked (True)-Generate route edges. The output Route Edges layer is populated with line features.   Unchecked (False)-Don't generate route edges. The output Route Edges layer is returned, but it is empty.

           populate_directions: Populate Directions (bool). Optional parameter.  Specify whether the tool should generate driving directions for
            each route.  Checked (True)-Indicates that the directions will be generated
            and configured based on the values for the Directions Language,
            Directions Style Name, and Directions Distance Units
            parameters.  Unchecked (False)-Directions are not generated, and the tool
            returns an empty Directions layer.

           directions_language: Directions Language (str). Optional parameter.  Specify the language that should be used when generating
            travel directions. This parameter is used only when the Populate
            Directions parameter is checked, or set to True. The parameter value can be
            specified using one of the following two- or five-character language codes:  ar-Arabic  bs-Bosnian ca-Catalancs-Czech da-Danish de-German  el-Greek en-English  es-Spanish et-Estonian fi-Finnish fr-French  he-Hebrew  hi-Hindi hr-Croatian hu-Hungarian id-Indonesian it-Italian  ja-Japanese  ko-Korean  lt-Lithuanian lv-Latvian nb-Norwegian nl-Dutch  pl-Polish  pt-BR-Brazilian
            Portuguese  pt-PT-European
            Portuguese  ro-Romanian ru-Russian  sl-Slovenian sr-Serbian sv-Swedish  th-Thaitr-Turkish vi-Vietnamese zh-CN-Simplified
            Chinese  zh-HK-Traditional Chinese (Hong Kong)
             zh-TW-Traditional Chinese (Taiwan)
            The tool first tries to find an exact match for the specified language including any language localization. If an exact match is not found, it tries to match the language family. If a match is still not found, the tool returns the directions using the default language, English. For example, if the directions language is specified as  es-MX (Mexican Spanish), the tool will return the directions in Spanish as it supports es language code and not es-MX.If a language supports localization, such as Brazilian Portuguese (pt-BR) and European Portuguese (pt-PT), you should specify the language family and   the localization. If you only specify the language family, the tool will not match the language family and instead return directions in the default language, English. For example, if the directions language is specified as pt, the tool will return the directions in English since it cannot decide if the directions should be returned in pt-BR or pt-PT.

           directions_distance_units: Directions Distance Units (str). Optional parameter.  Specify the units for displaying travel distance in the
            driving directions. This parameter is used only when the Populate
            Directions parameter is checked, or set to True.MilesKilometersMetersFeetYardsNauticalMiles
                  Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'NauticalMiles']

               directions_style_name: Directions Style Name (str). Optional parameter.  Specify the name of the formatting style for the
            directions. This parameter is used only when the Populate Directions parameter is checked, or set to True. The parameter can be specified
            using the following values:  NA Desktop-Generates turn-by-turn directions suitable
            for printing.  NA Navigation-Generates turn-by-turn directions designed
            for an in-vehicle navigation device.
                  Choice list:['NA Desktop', 'NA Navigation']

           travel_mode: Travel Mode (str). Optional parameter.  Specify the mode of transportation to model in the analysis. Travel modes are managed in ArcGIS Online and can be configured by the administrator of your organization to better reflect your organization's workflows. You need to specify the name of a travel mode supported by your organization. To get a list of supported travel mode names, use the same GIS server connection you used to access this tool, and from the Utilities toolbox, run GetTravelModes. The GetTravelModes tool adds a table, Supported Travel Modes, to the application. Any value in the Travel Mode Name field from the Supported Travel Modes table can be specified as input. You can also specify the value from the Travel Mode Settings field as input. This speeds up the tool execution because the tool does not have to look up the settings based on the travel mode name. The default value, Custom, allows you to configure your own travel mode using the custom travel mode parameters (UTurn at Junctions, Use Hierarchy, Restrictions, Attribute Parameter Values,  and Impedance). The default values of the custom travel mode parameters model traveling by car. You may want to choose Custom and set the custom travel mode parameters listed above to model a pedestrian with a fast walking speed or a truck with a given height, weight, and cargo of certain hazardous materials. You may choose to do this to try out different settings to get the analysis results you want. Once you have identified the analysis settings, you should work with your organization's administrator and save these settings as part of a new or existing travel mode so that everyone in your organization can rerun the analysis with the same settings. By choosing Custom, the values you set for the custom travel mode parameters are included in the analysis. Specifying  another travel mode, as defined by your organization, causes any values you set for the custom travel mode parameters to be ignored; the tool overrides them with values from your specified travel mode.

           impedance: Impedance (str). Optional parameter.  Specify the
            impedance, which is a value that represents the effort or cost of traveling along road segments or on other parts of the transportation network. Travel time is an impedance; a car may take one minute to travel a mile along an empty road. Travel times can vary by travel mode-a pedestrian may take more than 20  minutes to walk the same mile, so it is important to choose the right impedance for the travel mode you are modeling. Travel distance can also be an impedance; the length of a road in kilometers can be thought of as impedance. Travel distance in this sense is the same for all modes-a kilometer for a pedestrian is also a kilometer for a car. (What may change is the pathways on which the different modes are allowed to travel, which affects distance between points, and this is modeled by travel mode settings.)The value you provide for this parameter is ignored unless Travel Mode is set to Custom, which is the default value.Choose from the following impedance values:TravelTime-takes advantage of historical and live traffic data and is good for modeling the time it takes automobiles to travel along roads at a specific time of the day using live traffic speed data where available. When using TravelTime, you can optionally specify the TravelTime::Vehicle Maximum Speed (km/h) attribute parameter to specify the physical limitation of the speed the vehicle is capable of traveling.Minutes-does not use live traffic data but uses the historical average speeds for automobiles.TruckTravelTime-takes advantage of historical and live traffic data, but caps the speed to the posted truck speed limit. This is good for modeling the time it takes for the trucks to travel along roads at a specific time. When using TruckTravelTime, you can optionally specify the TruckTravelTime::Vehicle Maximum Speed (km/h) attribute parameter to specify the physical limitation of the speed the truck is capable of traveling.TruckMinutes-does not use live traffic data but uses the smaller of the historical average speeds for automobiles and the posted speed limits for trucks.WalkTime-defaults to a speed of 5 km/hr on all roads and paths, but this can be configured through the WalkTime::Walking Speed (km/h) attribute parameter.Miles-Stores  length measurements along roads in miles and can be used for performing analysis based on shortest distance.Kilometers-Stores  length measurements along roads in kilometers and can be used for performing analysis based on shortest distance.TimeAt1KPH-defaults to a speed of 1 km/hr on all roads and paths. The speed cannot be changed using any attribute parameters.If you choose a time-based impedance, such as TravelTime, TruckTravelTime, Minutes, TruckMinutes, or WalkTime, the Measurement Units parameter must be set to a time-based value; if you choose a distance-based impedance such as  Miles, Kilometers, Measurement Units must be distance-based.Drive Time, Truck Time, Walk Time, and Travel Distance impedance values are no longer supported and will be removed in a future release.   If you use one of these values, the tool uses the value of the  Time Impedance parameter for time-based values or Distance Impedance parameter for distance-based values.
                  Choice list:['Drive Time', 'Truck Time', 'Walk Time', 'Travel Distance', 'Minutes', 'TravelTime', 'TimeAt1KPH', 'WalkTime', 'TruckMinutes', 'TruckTravelTime', 'Miles', 'Kilometers']

           time_zone_for_time_windows: Time Zone for Time Windows (str). Optional parameter.  Specifies the time zone for the time window values on stops. The time windows are specified as part of TimeWindowStart and TimeWindowEnd fields on stops. This parameter is applicable only when the Use Time Windows parameter is checked (or set to True).
             The parameter can be specified using one of the following values:Geographically Local-The time window values associated with the stops are in the time zone in which the stops are located. For example, if the stop is located in an area that follows Eastern Standard Time and has time window values of 8 AM and 10 AM, the time window values will be treated as 8 AM and 10 AM in Eastern Standard Time. This is the default value.UTC-The time window values associated with the stops are in the Coordinated Universal Time (UTC). For example, if the stop is located in an area that follows Eastern Standard Time and has time window values of 8 AM and 10 AM, the time window values will be treated as 12 PM and 2 PM Eastern Standard Time assuming the Eastern Standard Time is obeying the Daylight Saving Time. Specifying the time window values in UTC is useful if you do not know the time zone in which the stops are located or when you have stops in multiple time zones and you want all the time windows to start simultaneously.
                  Choice list:['Geographically Local', 'UTC']

           save_output_network_analysis_layer: Save Output Network Analysis Layer (bool). Optional parameter.  Specify if the tool should save the analysis settings as a network analysis layer file. You cannot directly work with this file even when you open the file in an ArcGIS Desktop application like ArcMap. It is meant to be sent to Esri Technical Support to diagnose the quality of results returned from the tool.
            Checked (True)-Save the network analysis layer file. The file is downloaded in a temporary directory on your machine. In ArcGIS Pro, the location of the downloaded file can be determined  by viewing the value for the Output Network Analysis Layer parameter in the entry corresponding to the tool execution in the Geoprocessing history of your Project. In ArcMap, the location of the file can be determined by accessing the Copy Location option in the shortcut menu on the Output Network Analysis Layer parameter in the entry corresponding to the tool execution in the Geoprocessing Results window.   Unchecked (False)-Do not save the network analysis layer file. This is the default.

           overrides: Overrides (str). Optional parameter.  Specify additional settings that can influence the behavior of the solver when finding solutions for the network analysis problems.
             The value for this parameter needs to be specified in JavaScript Object Notation (JSON). For example, a valid value is of the following form {"overrideSetting1" : "value1", "overrideSetting2" : "value2"}. The override setting name is always enclosed in double quotation marks. The values can be a number, Boolean, or a string. The default value for this parameter is no
            value, which indicates not to override any solver
            settings. Overrides are advanced settings that should be
            used only after careful analysis of the results obtained before and
            after applying the settings. A list of supported override settings
            for each solver and their acceptable values can be obtained by
            contacting Esri Technical Support.

           save_route_data: Save Route Data (bool). Optional parameter.  Choose whether the output includes a zip file
            that contains a file geodatabase holding the inputs and outputs of
            the analysis in a format that can be used to share route layers
            with ArcGIS Online or Portal for ArcGIS.Checked (True)-Save the route data as a zip file. The file is downloaded in a temporary directory on your machine. In ArcGIS Pro, the location of the downloaded file can be determined by viewing the value for the Output Route Data parameter in the entry corresponding to the tool execution in the Geoprocessing history of your Project. In ArcMap, the location of the file can be determined by accessing the Copy Location option in the shortcut menu on the Output Route Data  parameter in the entry corresponding to the tool execution in the Geoprocessing Results window.   Unchecked (False)-Do not save the route data. This is the default.

           time_impedance: Time Impedance (str). Optional parameter.  Specify the time-based impedance, which is a value that represents the travel time along road segments or on other parts of the transportation network.If the impedance for the travel mode, as specified using the impedance parameter, is time-based, the value for time_impedance and impedance parameters should be identical. Otherwise the service will return an error.
                  Choice list:['Minutes', 'TravelTime', 'TimeAt1KPH', 'WalkTime', 'TruckMinutes', 'TruckTravelTime']

           distance_impedance: Distance Impedance (str). Optional parameter.  Specify the distance-based impedance, which is a value that represents the travel distance along road segments or on other parts of the transportation network.If the impedance for the travel mode, as specified using the impedance parameter, is distance-based, the value for distance_impedance and impedance parameters should be identical. Otherwise the service will return an error.
                  Choice list:['Miles', 'Kilometers']

           output_format: Output Format (str). Optional parameter.  Specify the format in which the output features are created.
            Choose from the following formats:Feature Set: The output features are returned as feature classes and tables. This is the default. JSON File: The output features are returned as a compressed file containing the JSON representation of the outputs. When this option is specified, the output is a single file (with a .zip extension) that contains one or more JSON files (with a .json extension) for each of the outputs created by the service. GeoJSON File: The output features are returned as a compressed file containing the GeoJSON representation of the outputs. When this option is specified, the output is a single file (with a .zip extension) that contains one or more GeoJSON files (with a .geojson extension) for each of the outputs created by the service.When a file based output format, such as JSON File or GeoJSON File, is specified,  no outputs are added to the display since the applications like ArcMap or ArcGIS Pro  do not know how to draw the contents of the result file. Instead, the result file is downloaded in a temporary directory on your machine. In ArcGIS Pro, the location of the downloaded file can be determined by viewing the value for the Output Result File parameter in the entry corresponding to the tool execution in the Geoprocessing history of your Project. In ArcMap, the location of the file can be determined by accessing the Copy Location option in the shortcut menu on the Output Result File  parameter in the entry corresponding to the tool execution in the Geoprocessing Results window.
                  Choice list:['Feature Set', 'JSON File']

           gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


           future: Optional, If True, a future object will be returns and the process will not wait for the task to complete. The default is False, which means wait for results.


           Returns the following as a named tuple:
               solve_succeeded - Solve Succeeded as a bool
               output_routes - Output Routes as a FeatureSet
               output_route_edges - Output Route Edges as a FeatureSet
               output_directions - Output Directions as a FeatureSet
               output_stops - Output Stops as a FeatureSet
               output_network_analysis_layer - Output Network Analysis Layer as a DataFile
               output_route_data - Output Route Data as a DataFile
               output_result_file - Output Result File as a DataFile
    """



    if gis is None:
        gis = arcgis.env.active_gis
    url = gis.properties.helperServices.asyncRoute.url
    tbx = import_toolbox(url, gis=gis)
    defaults = dict(zip(tbx.find_routes.__annotations__.keys(),
                        tbx.find_routes.__defaults__))
    if stops is None:
        stops = default_stops

    if point_barriers is None:
        point_barriers = defaults['point_barriers']

    if line_barriers is None:
        line_barriers = defaults['line_barriers']

    if polygon_barriers is None:
        polygon_barriers = defaults['polygon_barriers']

    if restrictions is None:
        restrictions = defaults['restrictions']

    if attribute_parameter_values is None:
        attribute_parameter_values = defaults['attribute_parameter_values']

    if route_line_simplification_tolerance is None:
        route_line_simplification_tolerance = defaults['route_line_simplification_tolerance']


    job = tbx.find_routes(stops=stops,
                          measurement_units=measurement_units,
                          analysis_region=analysis_region,
                          reorder_stops_to_find_optimal_routes=reorder_stops_to_find_optimal_routes,
                          preserve_terminal_stops=preserve_terminal_stops,
                          return_to_start=return_to_start,
                          use_time_windows=use_time_windows,
                          time_of_day=time_of_day,
                          time_zone_for_time_of_day=time_zone_for_time_of_day,
                          uturn_at_junctions=uturn_at_junctions,
                          point_barriers=point_barriers,
                          line_barriers=line_barriers,
                          polygon_barriers=polygon_barriers,
                          use_hierarchy=use_hierarchy,
                          restrictions=restrictions,
                          attribute_parameter_values=attribute_parameter_values,
                          route_shape=route_shape,
                          route_line_simplification_tolerance=route_line_simplification_tolerance,
                          populate_route_edges=populate_route_edges,
                          populate_directions=populate_directions,
                          directions_language=directions_language,
                          directions_distance_units=directions_distance_units,
                          directions_style_name=directions_style_name,
                          travel_mode=travel_mode,
                          impedance=impedance,
                          time_zone_for_time_windows=time_zone_for_time_windows,
                          save_output_network_analysis_layer=save_output_na_layer,
                          overrides=overrides,
                          save_route_data=save_route_data,
                          time_impedance=time_impedance,
                          distance_impedance=distance_impedance,
                          output_format=output_format,
                          gis=gis,
                          future=True)
    if future:
        return job
    res = job.result()
    return res

find_routes.__annotations__ = {
    'stops': FeatureSet,
    'measurement_units': str,
    'analysis_region': str,
    'reorder_stops_to_find_optimal_routes': bool,
    'preserve_terminal_stops': str,
    'return_to_start': bool,
    'use_time_windows': bool,
    'time_of_day': datetime,
    'time_zone_for_time_of_day': str,
    'uturn_at_junctions': str,
    'point_barriers': FeatureSet,
    'line_barriers': FeatureSet,
    'polygon_barriers': FeatureSet,
    'use_hierarchy': bool,
    'restrictions': str,
    'attribute_parameter_values': FeatureSet,
    'route_shape': str,
    'route_line_simplification_tolerance': LinearUnit,
    'populate_route_edges': bool,
    'populate_directions': bool,
    'directions_language': str,
    'directions_distance_units': str,
    'directions_style_name': str,
    'travel_mode': str,
    'impedance': str,
    'return': tuple}


# This code is based on https://github.com/kcg2015/Vehicle-Detection-and-Tracking by kyle Guan
try:
    import numpy as np       
    import cv2
    from numpy import dot
    from scipy.linalg import inv, block_diag
    from scipy.optimize import linear_sum_assignment
    import matplotlib.pyplot as plt
    from collections import deque
except Exception as e:
    pass

class Tracker(): # class for Kalman Filter-based tracker
    def __init__(self):
        # Initialize parametes for tracker (history)
        self.trackid = 0  # tracker's id
        self.class_name = ''
        self.score = 0.0
        self.box = [] # list to store the coordinates for a bounding box 
        self.hits = 0 # number of detection matches
        self.lost_tracks = 0 # number of unmatched tracks (track loss)
        
        # Initialize parameters for Kalman Filtering
        # The state is the (x, y) coordinates of the detection box
        # state: [up, up_dot, left, left_dot, down, down_dot, right, right_dot]
        # or[up, up_dot, left, left_dot, height, height_dot, width, width_dot]
        self.x_state = [] 
        self.dt = 1.   # time interval
        
        # Process matrix, assuming constant velocity model
        self.A = np.array([[1, self.dt, 0,  0,  0,  0,  0, 0],
                           [0, 1,  0,  0,  0,  0,  0, 0],
                           [0, 0,  1,  self.dt, 0,  0,  0, 0],
                           [0, 0,  0,  1,  0,  0,  0, 0],
                           [0, 0,  0,  0,  1,  self.dt, 0, 0],
                           [0, 0,  0,  0,  0,  1,  0, 0],
                           [0, 0,  0,  0,  0,  0,  1, self.dt],
                           [0, 0,  0,  0,  0,  0,  0,  1]])
        
        # Measurement matrix, assuming we can only measure the coordinates
        
        self.H = np.array([[1, 0, 0, 0, 0, 0, 0, 0],
                           [0, 0, 1, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 1, 0, 0, 0], 
                           [0, 0, 0, 0, 0, 0, 1, 0]])
        
        
        # Initialize the state covariance
        self.state_variance = 10.0
        self.P = np.diag(self.state_variance*np.ones(8))
        
        
        # Initialize the process covariance
        self.Q_comp_mat = np.array([[self.dt**4/4., self.dt**3/2.],
                                    [self.dt**3/2., self.dt**2]])
        self.Q = block_diag(self.Q_comp_mat, self.Q_comp_mat, 
                            self.Q_comp_mat, self.Q_comp_mat)
        
        # Initialize the measurement covariance
        self.R_scaler = 1.0
        self.R_diag_array = self.R_scaler * np.array([self.state_variance, self.state_variance, self.state_variance, self.state_variance])
        self.R = np.diag(self.R_diag_array)
   
    def kalman_filter(self, z): 
        '''
        Implement the Kalman Filter, including the predict and the update stages,
        with the measurement z
        '''
        x = self.x_state
        # Predict
        x = dot(self.A, x)
        self.P = dot(self.A, self.P).dot(self.A.T) + self.Q

        #Update
        S = dot(self.H, self.P).dot(self.H.T) + self.R
        K = dot(self.P, self.H.T).dot(inv(S)) # Kalman gain
        y = z - dot(self.H, x) # residual
        x += dot(K, y)
        self.P = self.P - dot(K, self.H).dot(self.P)
        self.x_state = x.astype(int) # convert to integer coordinates 
                                     #(pixel values)
        
    def predict(self):  
        '''
        Implment only the predict stage. This is used for unmatched detections and 
        unmatched tracks
        '''
        x = self.x_state
        x = dot(self.A, x)
        self.P = dot(self.A, self.P).dot(self.A.T) + self.Q
        self.x_state = x.astype(int)

def delete_trackers(deleted_tracks, tracker_list):
    '''
    Delete unused tracks from memory.

    '''
    for trk in deleted_tracks:
        tracker_list.remove(trk)

def box_iou(a, b):
    '''
    Helper funciton to calculate the ratio between intersection and the union of
    two boxes a and b
    a[0], a[1], a[2], a[3] <-> left, up, right, bottom
    '''
    w_intsec = np.maximum (0, (np.minimum(a[2], b[2]) - np.maximum(a[0], b[0])))
    h_intsec = np.maximum (0, (np.minimum(a[3], b[3]) - np.maximum(a[1], b[1])))
    s_intsec = w_intsec * h_intsec
    s_a = (a[2] - a[0])*(a[3] - a[1])
    s_b = (b[2] - b[0])*(b[3] - b[1])
  
    return float(s_intsec)/(s_a + s_b -s_intsec)

def munkres_assignment(trackers, detections, iou_thrd):
    '''
    From current list of trackers and new detections, output matched detections,
    unmatchted trackers, unmatched detections.
    '''    
    IOU_mat= np.zeros((len(trackers),len(detections)),dtype=np.float32)

    for t,trk in enumerate(trackers):
        for d,det in enumerate(detections):
            IOU_mat[t,d] = box_iou(trk,det) 
    
    # Produces matches       
    # Solve the maximizing the sum of IOU assignment problem using the
    # Hungarian algorithm (also known as Munkres algorithm)
    
    matched_idx_row, matched_idx_col = linear_sum_assignment(-IOU_mat)     

    unmatched_trackers, unmatched_detections = [], []
    for t,trk in enumerate(trackers):
        if(t not in matched_idx_row):
            unmatched_trackers.append(t)

    for d, det in enumerate(detections):
        if(d not in matched_idx_col):
            unmatched_detections.append(d)

    matches = []
   
    # For creating trackers we consider any detection with an 
    # overlap less than iou_thrd to signifiy the existence of 
    # an untracked object
    matched_idx = np.array([matched_idx_row,matched_idx_col]).T
    
    for m in matched_idx:
        if(IOU_mat[m[0],m[1]]<iou_thrd):
            unmatched_trackers.append(m[0])
            unmatched_detections.append(m[1])
        else:
            matches.append(m.reshape(1,2))
    
    if(len(matches)==0):
        matches = np.empty((0,2),dtype=int)
    else:
        matches = np.concatenate(matches,axis=0)
    
    return matches, np.array(unmatched_detections), np.array(unmatched_trackers)    


def main_tracker(img, detections, scores, assignment_iou_thrd, vanish_frames, detect_frames, tracker_list, tracker_ind):
    '''
    main_tracker function for detection and tracking
    '''
    vanish_frames = vanish_frames  # no.of consecutive unmatched detection before

    detect_frames = detect_frames  # no. of consecutive matches needed to establish a track

    x_box =[]
    z_box = detections # measurement
    
    if len(tracker_list) > 0:
        for trk in tracker_list:
            x_box.append(trk.box)
    
    # Hungarian/Munkres Assignment
    matched, unmatched_dets, unmatched_trks = munkres_assignment(x_box, z_box, iou_thrd = assignment_iou_thrd)  
    
    # Deal with matched detections     
    if matched.size >0:
        for trk_idx, det_idx in matched:
            z = z_box[det_idx]
            z = np.expand_dims(z, axis=0).T
            tmp_trk= tracker_list[trk_idx]
            tmp_trk.kalman_filter(z)
            xx = tmp_trk.x_state.T[0].tolist()
            xx =[xx[0], xx[2], xx[4], xx[6]]
            x_box[trk_idx] = xx
            tmp_trk.box =xx
            tmp_trk.score = scores[det_idx]
            tmp_trk.hits += 1
            tmp_trk.lost_tracks = 0
    
    # Deal with unmatched detections      
    if len(unmatched_dets)>0:
        for idx in unmatched_dets:
            z = z_box[idx]
            z = np.expand_dims(z, axis=0).T
            tmp_trk = Tracker() # Create a new tracker
            x = np.array([[z[0], 0, z[1], 0, z[2], 0, z[3], 0]]).T
            tmp_trk.x_state = x
            tmp_trk.predict()
            xx = tmp_trk.x_state
            xx = xx.T[0].tolist()
            xx =[xx[0], xx[2], xx[4], xx[6]]
            tmp_trk.box = xx
            tmp_trk.score = scores[idx]
            tmp_trk.trackid = tracker_ind # assign an ID for the tracker
            tracker_ind+=1
            tracker_list.append(tmp_trk)
            x_box.append(xx)
    
    # Deal with unmatched tracks       
    if len(unmatched_trks)>0:
        for trk_idx in unmatched_trks:
            tmp_trk = tracker_list[trk_idx]
            tmp_trk.lost_tracks += 1
            tmp_trk.predict()
            xx = tmp_trk.x_state
            xx = xx.T[0].tolist()
            xx =[xx[0], xx[2], xx[4], xx[6]]
            tmp_trk.box =xx
            x_box[trk_idx] = xx
                         
    # The list of tracks to be annotated  
    good_tracker_list =[]
    obj_info = {}
    predictions = []
    scores = []
    labels = []
    for trk in tracker_list:
        if ((trk.hits >= detect_frames) and (trk.lost_tracks <= vanish_frames)):
            good_tracker_list.append(trk)
            x_cv2 = trk.box
            obj_info[trk.trackid] = (x_cv2, trk.score)

            predictions.append([x_cv2[1], x_cv2[0], x_cv2[3] - x_cv2[1], x_cv2[2] - x_cv2[0]])
            scores.append(trk.score)
            labels.append(trk.trackid)

    # Book keeping
    deleted_tracks = filter(lambda x: x.lost_tracks > vanish_frames, tracker_list)  

    delete_trackers(deleted_tracks, tracker_list)

    return predictions, labels, scores, tracker_list, tracker_ind



from ._arcgis_model import ArcGISModel
import tempfile
from pathlib import Path
import json
from ._codetemplate import code
import random
import os, csv
import statistics
import warnings
from . import _tracker_util
from warnings import warn

HAS_OPENCV = True
HAS_FASTAI = True
HAS_ARCPY = True

# Try to import the necessary modules
# Exception will turn the HAS_FASTAI flag to false so that relevant exception can be raised
try:
    import torch
    import numpy as np
    import pandas as pd
    import PIL
    from fastai.vision.learner import create_body
    from fastprogress import progress_bar
    from fastai.vision import ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.vision.image import open_image, bb2hw, image2np, Image, pil2tensor
    from fastai.core import ifnone
    from torchvision import models
    from ._ssd_utils import SSDObjectCategoryList
    from ._retinanet_utils import RetinaNetModel, RetinaNetFocalLoss, compute_class_AP
    from .._data import prepare_data
    from fastai.callbacks import EarlyStoppingCallback
    from fastai.basic_train import Learner
    from .._data import _raise_fastai_import_error
    from ._arcgis_model import SaveModelCallback
    from .._image_utils import _get_image_chips, _get_transformed_predictions, _draw_predictions, _exclude_detection
    from .._video_utils import VideoUtils
except:
    HAS_FASTAI = False

try:
    import cv2
except:
    HAS_OPENCV = False

try:
    import arcpy
except:
    HAS_ARCPY = False

class RetinaNet(ArcGISModel):
    """
    Creates a RetinaNet Object Detector with the specified zoom scales
    and aspect ratios. 
    Based on the Fast.ai notebook at https://github.com/fastai/fastai_dev/blob/master/dev_nb/102a_coco.ipynb

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    scales                  Optional list of float values. Zoom scales of anchor boxes.
    ---------------------   -------------------------------------------
    ratios                  Optional list of float values. Aspect ratios of anchor
                            boxes.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `RetinaNet`, which
                            is `resnet50` by default. 
                            Compatible backbones: 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152'
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================
    
    :returns: `RetinaNet` Object
    """

    def __init__(self, data, scales=None, ratios=None, backbone=None, pretrained_path=None):

        # Set default backbone to be 'resnet50'
        if backbone is None: 
            backbone = models.resnet50

        super().__init__(data, backbone)

        # Check if a backbone provided is compatible, use resnet50 as default
        if not self._check_backbone_support(backbone):
            raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        self.name = "RetinaNet"
        self._code = code

        self.scales = ifnone(scales, [1,2**(-1/3), 2**(-2/3)])
        self.ratios = ifnone(ratios, [1/2,1,2])
        self._n_anchors = len(self.scales) * len(self.ratios)

        self._data = data
        self._chip_size = (data.chip_size,data.chip_size)

        # Cut-off the backbone before the penultimate layer
        self._encoder = create_body(self._backbone, -2)

        # Initialize the model, loss function and the Learner object        
        self._model = RetinaNetModel(self._encoder, n_classes=data.c-1, final_bias=-4, chip_size=self._chip_size, n_anchors=self._n_anchors)
        self._loss_f = RetinaNetFocalLoss(sizes=self._model.sizes, scales=self.scales, ratios=self.ratios)
        self.learn = Learner(data, self._model, loss_func=self._loss_f)
        self.learn.split([self._model.encoder[6], self._model.c5top5])
        self.learn.freeze()
        if pretrained_path is not None:
            self.load(str(pretrained_path))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    # Return a list of supported backbones names
    @property
    def supported_backbones(self):
        return [*self._resnet_family]

    def _create_emd(self, path):
        """
        Creates an Esri Model Definition (EMD) file with the parameters and 
        other information about the model.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. Path where the created 
                                Esri Model Definition file will be saved.
        =====================   ===========================================
        
        :returns: path of the saved EMD file
        """

        super()._create_emd(path)
        
        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
        self._emd_template["ModelConfiguration"] = "_RetinaNet_Inference"
        self._emd_template["ModelType"] = "ObjectDetection"
        self._emd_template["ExtractBands"] = [0, 1, 2]
        self._emd_template['ModelParameters']['scales'] = self._loss_f.scales #Scales and Ratios are attributes of RetinaNetFocalLoss object _loss_f
        self._emd_template['ModelParameters']['ratios'] = self._loss_f.ratios
        self._emd_template['Classes'] = []

        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]): # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        json.dump(self._emd_template, open(path.with_suffix('.emd'), 'w'), indent=4)
        return path.stem

    @property
    def _model_metrics(self):
        return {'accuracy': self.average_precision_score(show_progress=False)}

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a RetinaNet Object Detector from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        
        :returns: `RetinaNet` Object
        """

        emd_path = Path(emd_path)
        emd = json.load(open(emd_path))
        model_file = Path(emd['ModelFile'])
        chip_size = emd["ImageWidth"]

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
        
        resize_to = emd.get('resize_to')
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        data_passed = True
        # Create an image databunch for when loading the model using emd (without training data)
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                
                sd = ImageList([], path=tempfile.TemporaryDirectory().name).split_by_idx([])
                tempdata = sd.label_const(0, label_cls=SSDObjectCategoryList, classes=list(class_mapping.values())).transform(ds_tfms).databunch().normalize(imagenet_stats)
                tempdata.chip_size = chip_size
                tempdata.class_mapping = class_mapping
                tempdata.classes = ['background'] + list(class_mapping.values())
                data = tempdata
                data.c += 1 # Add 1 for background class

        data.resize_to = resize_to
        ret = cls(data, **emd['ModelParameters'], pretrained_path=model_file)
        
        if not data_passed:
            ret.learn.data.single_ds.classes = ret._data.classes
            ret.learn.data.single_ds.y.classes = ret._data.classes
        
        return ret


    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probabilty above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding 
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================
        
        """ 

        if rows > self._data.batch_size:
            rows = self._data.batch_size      
        self.learn.show_results(rows=rows, thresh=thresh, nms_overlap=nms_overlap, ssd=self)


    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            'assignment_iou_thrd': 0.3,
            'vanish_frames': 40,
            'detect_frames': 10
        },
        visual_options={
            'show_scores': True,
            'show_labels': True,
            'thickness': 2,
            'fontface': 0,
            'color': (255, 255, 255)
        }
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        =====================   ===========================================
        
        """

        VideoUtils.predict_video(
            self,
            input_video_path,
            metadata_file,
            threshold,
            nms_overlap,
            track, visualize,
            output_file_path,
            multiplex,
            multiplex_file_path,
            tracker_options,
            visual_options
        )

    def predict(self, image_path, threshold=0.5, nms_overlap=0.1, return_scores=True, visualize=False):
        """
        Predicts and displays the results of a trained model on a single image.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probabilty above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding 
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean.
                                Will return the probability scores of the 
                                bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with 
                                predicted bounding boxes if True.
        =====================   ===========================================
        
        :returns: 'List' of xmin, ymin, width, height of predicted bounding boxes on the given image
        """

        if not HAS_OPENCV:
            raise Exception("This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24")

        if isinstance(image_path, str):
            image = cv2.imread(image_path)
        else:
            image = image_path

        orig_height, orig_width, _ = image.shape
        orig_frame = image.copy()

        if self._data.resize_to is not None:
            if isinstance(self._data.resize_to, tuple):
                image = cv2.resize(image, self._data.resize_to)
            else:
                image = cv2.resize(image, (self._data.resize_to, self._data.resize_to))

        height, width, _ = image.shape

        if self._data.chip_size is not None:
            chips = _get_image_chips(image, self._data.chip_size)
        else:
            chips = [{'width': width, 'height': height, 'xmin': 0, 'ymin': 0, 'chip': image, 'predictions': []}]

        valid_tfms = self._data.valid_ds.tfms
        self._data.valid_ds.tfms = []

        for chip in chips:
            frame = Image(pil2tensor(PIL.Image.fromarray(cv2.cvtColor(chip['chip'], cv2.COLOR_BGR2RGB)), dtype=np.float32).div_(255))
            bbox = self.learn.predict(frame, thresh=threshold, nms_overlap=nms_overlap, ret_scores=True, ssd=self)[0]
            if bbox:
                scores = bbox.scores
                bboxes, lbls = bbox._compute_boxes()
                bboxes.add_(1).mul_(
                    torch.tensor([chip['height'] / 2, chip['width'] / 2, chip['height'] / 2, chip['width'] / 2])).long()
                for index, bbox in enumerate(bboxes):
                    if lbls is not None:
                        label = lbls[index]
                    else:
                        label = 'Default'

                    data = bb2hw(bbox)
                    if not _exclude_detection((data[0], data[1], data[2], data[3]), chip['width'], chip['height']):
                        chip['predictions'].append({
                            'xmin': data[0],
                            'ymin': data[1],
                            'width': data[2],
                            'height': data[3],
                            'score': float(scores[index]),
                            'label': label
                        })

        self._data.valid_ds.tfms = valid_tfms

        predictions, labels, scores = _get_transformed_predictions(chips)

        # Scale the predictions to original image and clip the predictions to image dims
        y_ratio = orig_height/height
        x_ratio = orig_width/width
        for index, prediction in enumerate(predictions):
            prediction[0] = prediction[0]*x_ratio
            prediction[1] = prediction[1]*y_ratio
            prediction[2] = prediction[2]*x_ratio
            prediction[3] = prediction[3]*y_ratio

            # Clip xmin
            if prediction[0] < 0: 
                prediction[2] = prediction[2] + prediction[0]
                prediction[0] = 1

            # Clip width when xmax greater than original width
            if prediction[0] + prediction[2] > orig_width:
                prediction[2] = (prediction[0] + prediction[2]) - orig_width

            # Clip ymin
            if prediction[1] < 0:
                prediction[3] = prediction[3] + prediction[1]
                prediction[1] = 1

            # Clip height when ymax greater than original height
            if prediction[1] + prediction[3] > orig_height:
                prediction[3] = (prediction[1] + prediction[3]) - orig_height

            predictions[index] = [
                prediction[0],
                prediction[1],
                prediction[2],
                prediction[3]
            ]

        if visualize:
            image = _draw_predictions(orig_frame, predictions, labels)
            import matplotlib.pyplot as plt
            plt.xticks([])
            plt.yticks([])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return predictions, labels, scores
        else:
            return predictions, labels

    def average_precision_score(self, detect_thresh=0.5, iou_thresh=0.1, mean=False, show_progress=True):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probabilty above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.                        
        =====================   ===========================================
        
        :returns: `dict` if mean is False otherwise `float`
        """

        aps = compute_class_AP(self, self._data.valid_dl, self._data.c - 1, show_progress, detect_thresh=detect_thresh, iou_thresh=iou_thresh)
        if mean:
            return statistics.mean(aps)
        else:
            return dict(zip(self._data.classes[1:], aps))